#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2006 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool traverser metaconfiguration code.
"""
from zope.publisher.interfaces.browser import IDefaultBrowserLayer
from zope.publisher.interfaces import IPublishTraverse
from zope.component.zcml import adapter as handle_adapter
from zope.component.zcml import view

from schooltool.traverser.traverser import PluggableTraverser
from schooltool.traverser.traverser import NullTraverserPlugin
from schooltool.traverser.traverser import AttributeTraverserPlugin
from schooltool.traverser.traverser import AdapterTraverserPlugin
from schooltool.traverser.interfaces import ITraverserPlugin


def traverser(_context, for_, type, factory,
              provides=IPublishTraverse,
              permission=None):
    view(_context, [factory], type, '', [for_],
         permission=permission,
         allowed_interface=None, allowed_attributes=None,
         provides=provides)


def pluggableTraverser(_context, for_, type,
                       factory=PluggableTraverser,
                       provides=IPublishTraverse,
                       permission=None):
    view(_context, [factory], type, '', [for_],
         permission=permission,
         allowed_interface=None, allowed_attributes=None,
         provides=provides)


def traverserPlugin(_context, for_, plugin,
                    name='',
                    layer=IDefaultBrowserLayer,
                    permission=None):
    plugin = type(plugin.__name__, (plugin, ), {'__name__': name})
    handle_adapter(_context, [plugin],
                   provides=ITraverserPlugin,
                   for_=(for_, layer),
                   permission=permission,
                   name=name)


def nullTraverserPlugin(_context, for_, name,
                        layer=IDefaultBrowserLayer,
                        permission=None):
    traverserPlugin(
        _context, for_, NullTraverserPlugin,
        name=name, layer=layer, permission=permission)


def attributeTraverserPlugin(_context, for_, name,
                             layer=IDefaultBrowserLayer,
                             permission=None):
    traverserPlugin(
        _context, for_, AttributeTraverserPlugin,
        name=name, layer=layer, permission=permission)


def adapterTraverserPlugin(_context, for_, name, adapter,
                           adapter_name='',
                           layer=IDefaultBrowserLayer,
                           permission=None):
    plugin = type('%sAdapterTraverserPlugin' % adapter.__name__,
                  (AdapterTraverserPlugin,),
                  {'adapterName': adapter_name,
                   'interface': adapter})
    traverserPlugin(_context, for_, plugin,
                    name=name, layer=layer, permission=permission)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Selenium functional tests setup for schooltool.basicperson
"""

import os

from schooltool.testing.selenium import SeleniumLayer

dir = os.path.abspath(os.path.dirname(__file__))
filename = os.path.join(dir, 'stesting.zcml')

basicperson_selenium_layer = SeleniumLayer(filename,
                                           __name__,
                                           'basicperson_selenium_layer')

def registerSeleniumSetup():
    try:
        import selenium
    except ImportError:
        return
    from schooltool.testing import registry
    import schooltool.testing.selenium

    def login(browser, username, password):
        browser.open('http://localhost')
        browser.query.link('Log in').click()
        browser.query.name('username').type(username)
        browser.query.name('password').type(password)
        page = browser.query.tag('html')
        browser.query.button('Log in').click()
        browser.wait(lambda: page.expired)

    registry.register('SeleniumHelpers',
        lambda: schooltool.testing.selenium.registerBrowserUI(
            'login', login))

    def addPerson(browser, first_name, last_name, username, password, **kw):
        optional = (
            'prefix',
            'middle_name',
            'suffix',
            'preferred_name',
            'gender',
            'birth_date',
            'ID',
            'ethnicity',
            'language',
            'placeofbirth',
            'citizenship',
            'group',
            'advisor',
            )
        browser.open('http://localhost/persons')
        page = browser.query.tag('html')
        browser.query.link('Person').click()
        browser.wait(lambda: page.expired)
        browser.query.name('form.widgets.first_name').type(first_name)
        browser.query.name('form.widgets.last_name').type(last_name)
        browser.query.name('form.widgets.username').type(username)
        browser.query.name('form.widgets.password').type(password)
        browser.query.name('form.widgets.confirm').type(password)
        for name in optional:
            if name in kw:
                value = kw[name]
                widget_id = ''.join(['form-widgets-', name])
                browser.query.id(widget_id).ui.set_value(value)
        page = browser.query.tag('html')
        browser.query.button('Submit').click()
        browser.wait(lambda: page.expired)

    registry.register('SeleniumHelpers',
        lambda: schooltool.testing.selenium.registerBrowserUI(
            'person.add', addPerson))

    def addAdvisors(browser, username, advisors):
        browser.open('http://localhost/persons/%s/advisors.html' % username)
        selector = 'available_table-ajax-available_table--title'
        browser.query.id(selector).type(', '.join(advisors))
        selector = '#available_table-ajax-available_table- table'
        table = browser.query.css(selector)
        browser.query.name('SEARCH_BUTTON').click()
        browser.wait(lambda: table.expired)
        # XXX: Click Show All here in case there are lots of people
        selector = '#available_table-ajax-available_table- table'
        table = browser.query.css(selector)
        browser.query.name('ADD_DISPLAYED_RESULTS').click()
        browser.wait(lambda: table.expired)

    registry.register('SeleniumHelpers',
        lambda: schooltool.testing.selenium.registerBrowserUI(
            'person.advisors.add', addAdvisors))

    def addAdvisees(browser, username, advisees):
        browser.open('http://localhost/persons/%s/advisees.html' % username)
        selector = 'available_table-ajax-available_table--title'
        browser.query.id(selector).type(', '.join(advisees))
        selector = '#available_table-ajax-available_table- table'
        table = browser.query.css(selector)
        browser.query.name('SEARCH_BUTTON').click()
        browser.wait(lambda: table.expired)
        # XXX: Click Show All here in case there are lots of people
        selector = '#available_table-ajax-available_table- table'
        table = browser.query.css(selector)
        browser.query.name('ADD_DISPLAYED_RESULTS').click()
        browser.wait(lambda: table.expired)

    registry.register('SeleniumHelpers',
        lambda: schooltool.testing.selenium.registerBrowserUI(
            'person.advisees.add', addAdvisees))

registerSeleniumSetup()
del registerSeleniumSetup



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.relationship.uri
"""
import unittest
import doctest


def doctest_URIObject():
    """Tests for URIObject.

    URIObject's constructor takes three arguments: URI, name and description

        >>> from schooltool.relationship.uri import URIObject
        >>> uri = URIObject('http://example.com', 'Example', 'An example.')
        >>> uri
        <URIObject Example>
        >>> uri.uri
        'http://example.com'
        >>> uri.name
        'Example'
        >>> uri.description
        'An example.'

    The attributes of an URIObject match those defined by IURIObject

        >>> from zope.interface.verify import verifyObject
        >>> from schooltool.relationship.uri import IURIObject
        >>> verifyObject(IURIObject, uri)
        True

    Description is optional

        >>> uri = URIObject('http://example.com', 'Example')
        >>> uri.description
        ''

    Name is also optional

        >>> uri = URIObject('http://example.com')
        >>> uri.name

    XXX Why does description default to '', while name defaults to None?

    URIs must be syntactically valid

        >>> URIObject('not a URI')
        Traceback (most recent call last):
          ...
        ValueError: This does not look like a URI: 'not a URI'

    URIObjects are comparable.  Two URIObjects are equal if, and only if,
    their 'uri' attributes are equal.

        >>> uri2 = URIObject('http://example.com', 'Exampleur')
        >>> uri3 = URIObject('http://example.org', 'Exampleur')
        >>> uri2 == uri
        True
        >>> uri2 == uri3
        False
        >>> uri2 != uri
        False
        >>> uri2 != uri3
        True
        >>> uri == 'example:Just a string'
        False
        >>> uri != 'example:Just a string'
        True

    By the way, there are separate tests for == and != because in Python
    __eq__ and __ne__ are two different methods.

    What should uri == 'http://example.com' return?  I have decided that
    it should return False, because applications may rely on roles having
    `name` and `description` attributes, after they check the relationship
    type.

        >>> uri == 'http://example.com'
        False
        >>> uri != 'http://example.com'
        True

    URIObjects are hashable.  Equal objects must hash to the same value

        >>> hash(uri) == hash(uri2)
        True

    URIObjects are immutable

        >>> uri.uri = 'http://makemoneyfast.example.net'
        Traceback (most recent call last):
          ...
        AttributeError: can't set attribute

        >>> uri.name = 'Friendship'
        Traceback (most recent call last):
          ...
        AttributeError: can't set attribute

        >>> uri.description = 'Dunno'
        Traceback (most recent call last):
          ...
        AttributeError: can't set attribute

    """


def test_suite():
    return unittest.TestSuite([
                doctest.DocTestSuite('schooltool.relationship.uri'),
                doctest.DocTestSuite(),
           ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
URI objects.

Relationship types and roles are identified by URIs (the idea was borrowed
from XLink and RDF).  Instead of dealing with strings directly,
schooltool.relationship uses introspectable URI objects that also have an
optional short name and a description in addition to the URI itself.

By convention, names of global URI object constants start with 'URI'.
"""

import re

from zope.interface import Interface, implements
from zope.schema import Text, TextLine, URI


class IURIObject(Interface):
    """An opaque identifier of a role or a relationship type.

    Roles and relationships are identified by URIs in XML representation.
    URI objects let the application assign human-readable names to roles
    and relationship types.

    URI objects are equal iff their uri attributes are equal.

    URI objects are hashable.
    """

    uri = URI(title=u"URI",
            description=u"The URI (as a string).")

    name = TextLine(title=u"Name",
            description=u"Human-readable name.")

    description = Text(title=u"Description",
            description=u"Human-readable description.")


class URIObject(object):
    """See IURIObject."""

    implements(IURIObject)

    def __init__(self, uri, name=None, description=''):
        if not looks_like_a_uri(uri):
            raise ValueError("This does not look like a URI: %r" % uri)
        self._uri = uri
        self._name = name
        self._description = description

    uri = property(lambda self: self._uri)
    name = property(lambda self: self._name)
    description = property(lambda self: self._description)

    def __eq__(self, other):
        if not IURIObject.providedBy(other):
            return False
        return self.uri == other.uri

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.uri)

    def __repr__(self):
        return '<URIObject %s>' % (self.name or self.uri)


def looks_like_a_uri(uri):
    r"""Check if the argument looks like a URI string.

    Refer to http://www.ietf.org/rfc/rfc2396.txt for details.
    We're only approximating to the spec.

    Some examples of valid URI strings:

        >>> looks_like_a_uri('http://foo/bar?baz#quux')
        True
        >>> looks_like_a_uri('HTTP://foo/bar?baz#quux')
        True
        >>> looks_like_a_uri('mailto:root')
        True

    These strings are all invalid URIs:

        >>> looks_like_a_uri('2HTTP://foo/bar?baz#quux')
        False
        >>> looks_like_a_uri('\nHTTP://foo/bar?baz#quux')
        False
        >>> looks_like_a_uri('mailto:postmaster ')
        False
        >>> looks_like_a_uri('mailto:postmaster text')
        False
        >>> looks_like_a_uri('nocolon')
        False
        >>> looks_like_a_uri(None)
        False

    """
    uri_re = re.compile(r"^[A-Za-z][A-Za-z0-9+-.]*:\S\S*$")
    return bool(uri and uri_re.match(uri) is not None)




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Email functionality

"""

import email.Charset
from email.MIMEText import MIMEText
from datetime import datetime
import pytz
import smtplib
import socket

from persistent import Persistent
from persistent.dict import PersistentDict
from zope.container.btree import BTreeContainer
from zope.container.contained import Contained
from zope.container.interfaces import INameChooser
from zope.component import adapter
from zope.interface import implements, implementer

from schooltool.app.app import InitBase, StartUpBase
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.common import SchoolToolMessage as _
from schooltool.email.interfaces import IEmailContained, IEmailContainer
from schooltool.email.interfaces import IEmailUtility


email.Charset.add_charset('utf-8', email.Charset.SHORTEST, None, None)
EMAIL_KEY = 'schooltool.email'


status_messages = {
    10: _('The SchoolTool mail service is disabled'),
    20: _("Couldn't connect to the SMTP server (${info})"),
    30: _('Error sending HELO to the SMTP server (${info})'),
    40: _("Couldn't login as ($username) to SMTP server (${info})"),
    50: _('The server (${info}) rejected the From address: ${from_address}'),
    60: _('The server (${info}) rejected the following recipient addresses: '
          '${addresses}'),
    70: _('The server (${info}) replied that the message data was malformed'),
    }


class Email(Persistent, Contained):

    implements(IEmailContained)

    def __init__(self, from_address, to_addresses, body, subject=None):
        self.from_address = from_address
        self.to_addresses = to_addresses
        self.body = body
        self.subject = subject
        self.status_code = None
        self.status_parameters = PersistentDict()
        self.time_created = pytz.utc.localize(datetime.utcnow())
        self.time_sent = None


class EmailContainer(BTreeContainer):

    implements(IEmailContainer)

    enabled = None
    hostname = None
    port = None
    username = None
    password = None
    tls = None


class EmailAppStartup(StartUpBase):

    def __call__(self):
        if EMAIL_KEY not in self.app:
            self.app[EMAIL_KEY] = EmailContainer()


class EmailInit(InitBase):

    def __call__(self):
        self.app[EMAIL_KEY] = EmailContainer()


@implementer(IEmailContainer)
@adapter(ISchoolToolApplication)
def getEmailContainer(app):
    return app.get(EMAIL_KEY)


class EmailUtility(object):

    implements(IEmailUtility)

    smtp_factory = None

    def getEmailContainer(self):
        app = ISchoolToolApplication(None)
        return IEmailContainer(app)

    def getApplicationTimezone(self):
        app = ISchoolToolApplication(None)
        return pytz.timezone(IApplicationPreferences(app).timezone)

    def emailAsString(self, email):
        message = MIMEText(email.body.encode('utf-8'), 'plain', 'utf-8')
        message['From'] = email.from_address.encode('utf-8')
        message['To'] = ', '.join([address.encode('utf-8')
                                   for address in email.to_addresses])
        if email.subject is not None:
            message['Subject'] = email.subject.encode('utf-8')
        time_format = '%a, %d %b %Y %H:%M:%S %z'
        application_timezone = self.getApplicationTimezone()
        time_created = email.time_created.astimezone(application_timezone)
        message['Date'] = time_created.strftime(time_format).encode('utf-8')
        return message.as_string()

    def queue(self, email, status_code=None, status_parameters={}):
        email.status_code = status_code
        email.status_parameters.update(status_parameters)
        email.time_sent = pytz.utc.localize(datetime.utcnow())
        if email.__parent__ is not self.container:
            name = INameChooser(self.container).chooseName('', email)
            self.container[name] = email

    def enabled(self):
        if self.smtp_factory is None:
            return False
        container = self.getEmailContainer()
        return container.enabled

    def send(self, email):
        self.container = self.getEmailContainer()
        if not self.enabled():
            self.queue(email, 10)
            return False
        server_info = '%s:%d' % (self.container.hostname,
                                 self.container.port or 25)
        try:
            connection = self.smtp_factory()
            if self.container.port:
                port = str(self.container.port)
            else:
                port = '25'
            connection.connect(self.container.hostname, port)
            code, response = connection.ehlo()
            if code < 200 or code >= 300:
                code, response = connection.helo()
                if code < 200 or code >= 300:
                    raise smtplib.SMTPHeloError(code, response)
            if connection.has_extn('starttls') and self.container.tls:
                connection.starttls()
                connection.ehlo()
            if connection.does_esmtp:
                if self.container.username is not None and \
                   self.container.password is not None:
                    connection.login(self.container.username,
                                     self.container.password)
        except (socket.error,), e:
            self.queue(email, 20, {'info': server_info})
            return False
        except (smtplib.SMTPHeloError,), e:
            self.queue(email, 30, {'info': server_info})
            connection.quit()
            return False
        except (smtplib.SMTPException,), e:
            self.queue(email, 40, {'info': server_info,
                                   'username': self.container.username})
            connection.quit()
            return False
        message = self.emailAsString(email)
        result = {}
        try:
            result = connection.sendmail(email.from_address,
                                         email.to_addresses,
                                         message)
        except (smtplib.SMTPSenderRefused,), e:
            self.queue(email, 50, {'info': server_info,
                                   'from_address': e.sender})
            connection.quit()
            return False
        except (smtplib.SMTPRecipientsRefused,), e:
            addresses = e.recipients.keys()
            self.queue(email, 60, {'info': server_info,
                                   'addresses': ', '.join(addresses)})
            connection.quit()
            return False
        except (smtplib.SMTPHeloError,), e:
            self.queue(email, 30, {'info': server_info})
            connection.quit()
            return False
        except (smtplib.SMTPDataError,), e:
            self.queue(email, 70, {'info': server_info})
            connection.quit()
            return False
        if result:
            addresses = [address for address in email.to_addresses
                         if address in result.keys()]
            email.to_addresses = addresses[:]
            self.queue(email, 60, {'info': server_info,
                                   'addresses': ', '.join(addresses)})
            connection.quit()
            return False
        connection.quit()
        return True


class SMTPEmailUtility(EmailUtility):

    smtp_factory = smtplib.SMTP



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for SchoolTool XLS export views.
"""
import itertools
import xlrd
import datetime


class Blank(str):
    def __new__(cls, show_text=''):
        return str.__new__(cls, show_text)
    def __init__(self, show_text=''):
        str.__init__(self, show_text)
    def __repr__(self):
        return str(self)


class NoCell(Blank):
    pass


class DateTime(datetime.datetime):
    def __repr__(self):
        if self.time().isoformat() == '00:00:00':
            return self.strftime('%Y-%m-%d')
        return self.strftime('%Y-%m-%d %H:%M:%S')


def format_cell(sheet, row, col):
    try:
        cell = sheet.cell(row, col)
    except IndexError:
        return NoCell()
    if cell.ctype == xlrd.XL_CELL_BLANK:
        # Contains formatting information but no data
        return Blank()
    elif cell.ctype == xlrd.XL_CELL_EMPTY:
        return Blank()
    elif cell.ctype == xlrd.XL_CELL_BOOLEAN:
        return bool(int(cell.value))
    elif cell.ctype == xlrd.XL_CELL_DATE:
        data = xlrd.xldate_as_tuple(cell.value, sheet.book.datemode)
        return DateTime(*data)
    elif cell.ctype == xlrd.XL_CELL_NUMBER:
        return float(cell.value)
    elif cell.ctype == xlrd.XL_CELL_TEXT:
        return unicode(cell.value)
    elif cell.ctype == xlrd.XL_CELL_ERROR:
        try:
            return NoCell(show_text=xlrd.error_text_from_code.get(int(cell.value), cell))
        except ValueError:
            return NoCell(show_text=repr(cell))


def col_name(n):
    assert n > 0
    name = ''
    while n:
        n = n-1
        name = chr(65+n%26)+name
        n = n/26
    return name


def col_n(name):
    try:
        return int(name)
    except ValueError:
        pass
    name = name.upper()
    i = 0
    while name:
        i = i*26+(ord(name[0])-64)
        name = name[1:]
    return i


def str_to_n_list(s):
    result = []
    ranges = s.split(',')
    for r in ranges:
        parts = [p.strip() for p in r.split('-')]
        if not parts:
            continue
        if len(parts) == 1:
            result.append(col_n(parts[0]))
        elif len(parts) == 2:
            result.extend(range(col_n(parts[0]), col_n(parts[1])+1))
        else:
            raise ValueError('%r is not a range' % r)
    return result


def list_rows_cols(sheet, rows, cols):
    if rows is None:
        rows = range(1, sheet.nrows+1)
    elif isinstance(rows, str):
        rows = str_to_n_list(rows)
    elif not hasattr(rows, '__iter__'):
        rows = rows,

    if cols is None:
        cols = range(1, sheet.ncols+1)
    elif isinstance(cols, str):
        cols = str_to_n_list(cols)
    elif not hasattr(cols, '__iter__'):
        cols = cols,

    rows = list(rows)
    cols = list(cols)
    return rows, cols


def build_sheet_table(sheet, rows=None, cols=None):
    rows, cols = list_rows_cols(sheet, rows, cols)

    table = []
    if not rows or not cols:
        return table

    trow = [NoCell(show_text='*')]
    prev_col = cols[0]-1
    for ncol in cols:
        if ncol != prev_col+1:
            trow.append(NoCell())
        trow.append(NoCell(show_text=col_name(ncol)))
        prev_col = ncol
    table.append(trow)

    prev_row = rows[0]-1
    for nrow in rows:
        assert nrow > 0
        if nrow != prev_row+1:
            table.append([NoCell(show_text='...')])
        trow = [NoCell(show_text=' %d ' % (nrow))]
        prev_col = cols[0]-1
        for ncol in cols:
            if ncol > prev_col+1:
                trow.append(NoCell(show_text='...'))
            prev_col = ncol
            cell = format_cell(sheet, nrow-1, ncol-1)
            trow.append(cell)
        prev_row = nrow
        table.append(trow)

    return table


def format_sheet_table(table):
    mlen = [max([len(repr(cell)) for cell in col])
            for col in itertools.izip_longest(*table, fillvalue=NoCell())]
    def val(cell, ncoll):
        r = repr(cell)
        l = len(r)
        s = (mlen[ncoll]-l)
        val = '%s%s%s' % (' '*(s/2), r, ' '*(s-s/2))
        if isinstance(cell, NoCell):
            return ' %s ' % val
        return '[%s]' % val
    for y, row in enumerate(table):
        for x, cell in enumerate(row):
            table[y][x] = val(cell, x)
    return '\n'.join([''.join([str(c) for c in r]) for r in table])


def print_sheet(sheet, rows=None, cols=None):
    table = build_sheet_table(sheet, rows=rows, cols=cols)
    print format_sheet_table(table)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool flourish skin.
"""

import ajax
import breadcrumbs
import containers
import content
import error
import fields
import form
import stesting
import interfaces
import page
import resource
import report
import sorting
import tal
import templates
import viewlet
import widgets

from .helpers import *

from schooltool.skin.flourish.interfaces import IFlourishLayer



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Term Interfaces
"""
__docformat__ = 'reStructuredText'

import zope.schema
from zope.container.constraints import contains, containers
from zope.container.interfaces import IContainer
from zope.location.interfaces import ILocation, IContained
from zope.interface import Interface, Attribute

from schooltool.common import SchoolToolMessage as _
from schooltool.common import IDateRange


class ITerm(IDateRange, IContained):
    """A term is a set of school days inside a given date range."""
    containers('.ITermContainer')

    __name__ = zope.schema.TextLine(
        title=_("SchoolTool ID"),
        description=_(
            """An internal identifier of this term."""),
        required=True)

    title = zope.schema.TextLine(
        title=_("Title"))

    def isSchoolday(date):
        """Return whether the date is a schoolday.

        Raises a ValueError if the date is outside of the term covered.
        """


class ITermWrite(Interface):
    """A term is a set of school days inside a given date range.

    This interface defines an term that can be modified.
    """

    def add(day):
        """Mark the day as a schoolday.

        Raises a ValueError if the date is outside of the term covered.
        """

    def remove(day):
        """Mark the day as a holiday.

        Raises a ValueError if the date is outside of the term covered.
        """

    def reset(first, last):
        """Change the term and mark all days as holidays.

        If first is later than last, a ValueError is raised.
        """

    def addWeekdays(*weekdays):
        """Mark that all days of week with a number in weekdays within the
        term will be schooldays.

        The numbering used is the same as one used by datetime.date.weekday()
        method, or the calendar module: 0 is Monday, 1 is Tuesday, etc.
        """

    def removeWeekdays(*weekdays):
        """Mark that all days of week with a number in weekdays within the
        term will be holidays.

        The numbering used is the same as one used by datetime.date.weekday()
        method, or the calendar module: 0 is Monday, 1 is Tuesday, etc.
        """

    def toggleWeekdays(*weekdays):
        """Toggle the state of all days of week with a number in weekdays.

        The numbering used is the same as one used by datetime.date.weekday()
        method, or the calendar module: 0 is Monday, 1 is Tuesday, etc.
        """


class ITermContainer(IContainer, ILocation):
    """A container for terms.

    It stores term calendars for registered term IDs.
    """
    contains(ITerm)


class IDateManager(Interface):
    """A class that handles dates and time.

    It does so taking the preferred timezone into account.
    """

    today = zope.schema.Date(
        title=u"Today",
        description=u"""The current day.""",
        required=True)

    current_term = Attribute("The active term.")


class TermDateNotInSchoolYear(Exception):

    def __repr__(self):
        return "Dates do not fit in school year!"

    __str__ = __repr__





#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for schooltool.app.browser.containers.
"""
import unittest
import doctest

from zope.app.testing import setup
from zope.interface import directlyProvides
from zope.publisher.browser import TestRequest

from schooltool.app.browser.testing import setUp, tearDown


def doctest_ContainerView():
    r"""Test for ContainerView

    Let's create some persons to toy with in a person container:

        >>> from schooltool.skin.containers import ContainerView
        >>> from schooltool.person.person import Person, PersonContainer
        >>> setup.setUpAnnotations()

        >>> personContainer = PersonContainer()
        >>> from zope.traversing.interfaces import IContainmentRoot
        >>> directlyProvides(personContainer, IContainmentRoot)

        >>> personContainer['pete'] = Person('pete', 'Pete Parrot')
        >>> personContainer['john'] = Person('john', 'Long John')
        >>> personContainer['frog'] = Person('frog', 'Frog Man')
        >>> personContainer['toad'] = Person('toad', 'Taodsworth')
        >>> request = TestRequest()
        >>> view = ContainerView(personContainer, request)

    After calling update, we should have a batch setup with everyone in it:

        >>> view.update()
        >>> [p.title for p in view.batch]
        ['Frog Man', 'Long John', 'Pete Parrot', 'Taodsworth']


    We can alter the batch size and starting point through the request

        >>> request.form = {'batch_start': '2',
        ...                 'batch_size': '2'}
        >>> view.update()
        >>> [p.title for p in view.batch]
        ['Pete Parrot', 'Taodsworth']

    We can search through the request:

        >>> request.form = {'SEARCH': 'frog'}
        >>> view.update()
        >>> [p.title for p in view.batch]
        ['Frog Man']

    And we can clear the search (which ignores any search value):

        >>> request.form = {'SEARCH': 'frog',
        ...                 'CLEAR_SEARCH': 'on'}
        >>> view.update()
        >>> [p.title for p in view.batch]
        ['Frog Man', 'Long John', 'Pete Parrot', 'Taodsworth']

    """


def doctest_ContainerDeleteView():
    r"""Test for ContainerDeleteView

    Let's create some persons to delete from a person container:

        >>> from schooltool.skin.containers import ContainerDeleteView
        >>> from schooltool.person.person import Person, PersonContainer
        >>> setup.setUpAnnotations()

        >>> personContainer = PersonContainer()

        >>> from zope.traversing.interfaces import IContainmentRoot
        >>> directlyProvides(personContainer, IContainmentRoot)

        >>> personContainer['pete'] = Person('pete', 'Pete Parrot')
        >>> personContainer['john'] = Person('john', 'Long John')
        >>> personContainer['frog'] = Person('frog', 'Frog Man')
        >>> personContainer['toad'] = Person('toad', 'Taodsworth')
        >>> request = TestRequest()
        >>> view = ContainerDeleteView(personContainer, request)

    We should have the list of all the Ids of items that are going to
    be deleted from container:

        >>> view.listIdsForDeletion()
        []

    We must pass ids of selected people in the request:

        >>> request.form = {'delete.pete': 'on',
        ...                 'delete.john': 'on',
        ...                 'CONFIRM': 'Confirm'}
        >>> ids = [key for key in view.listIdsForDeletion()]
        >>> ids.sort()
        >>> ids
        [u'john', u'pete']
        >>> [item.title for item in view.itemsToDelete]
        ['Long John', 'Pete Parrot']

    These two should be gone after update:

        >>> view.update()
        >>> ids = [key for key in personContainer]
        >>> ids.sort()
        >>> ids
        [u'frog', u'toad']

    And we should be redirected to the container view:

        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1'

    If we press Cancel no one should get hurt though:

        >>> request.form = {'delete.frog': 'on',
        ...                 'delete.toad': 'on',
        ...                 'CANCEL': 'Cancel'}

    You see, both our firends are still in there:

        >>> ids = [key for key in personContainer]
        >>> ids.sort()
        >>> ids
        [u'frog', u'toad']

    But we should be redirected to the container:

        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1'

    No redirection if nothing was pressed should happen:

        >>> request.form = {'delete.frog': 'on',
        ...                 'delete.toad': 'on'}
        >>> view.update()
        >>> request.response.getHeader('Location')
        'http://127.0.0.1'

    """


def doctest_TableContainerView():
    r"""Test for ContainerView

    Let's create some persons to toy with in a table container:

        >>> from schooltool.skin.containers import TableContainerView
        >>> from schooltool.person.person import Person, PersonContainer
        >>> setup.setUpAnnotations()

        >>> personContainer = PersonContainer()
        >>> from zope.traversing.interfaces import IContainmentRoot
        >>> directlyProvides(personContainer, IContainmentRoot)

        >>> personContainer['pete'] = Person('pete', 'Pete Parrot')
        >>> personContainer['john'] = Person('john', 'Long John')
        >>> personContainer['frog'] = Person('frog', 'Frog Man')
        >>> personContainer['toad'] = Person('toad', 'Taodsworth')
        >>> request = TestRequest()

    The table view is using the ITableFormatter adapter to lookup the
    approrpiate table formatter for the container:

        >>> from zope.component import provideAdapter
        >>> from schooltool.table.interfaces import ITableFormatter
        >>> from zope.interface import implements
        >>> class TableFormatter(object):
        ...     implements(ITableFormatter)
        ...     def __init__(self, context, request):
        ...         self.request = request
        ...     def setUp(self, **kwargs):
        ...         print "Setting up table formatter: %s" % kwargs

        >>> from schooltool.person.interfaces import IPersonContainer
        >>> from zope.publisher.interfaces.browser import IBrowserRequest
        >>> provideAdapter(TableFormatter, adapts=[IPersonContainer,
        ...                                        IBrowserRequest],
        ...                                provides=ITableFormatter)

    By default the template that displays a list of items is being
    used:

        >>> view = TableContainerView(personContainer, request)
        >>> view.canModify = lambda: False
        >>> view.template = lambda: "The table template."
        >>> view.delete_template = lambda: "The delete template."
        >>> result = view()
        Setting up table formatter:
         {'columns_before': [],
          'formatters': [<function url_cell_formatter at ...>]}

        >>> result
        'The table template.'

    If we can modify the list that is being displayed, an additional
    columns is added before the default columns:

        >>> view.canModify = lambda: True
        >>> result = view()
        Setting up table formatter:
            {'columns_before': [<....DependableCheckboxColumn object at ...>],
             'formatters': [<function url_cell_formatter at ...>]}

        >>> result
        'The table template.'


    If there 'DELETE' button is pressed, a different template is used:

        >>> request.form = {'delete.pete': 'on', 'DELETE': 'Delete'}
        >>> view()
        'The delete template.'

    We should have the list of all the Ids of items that are going to
    be deleted from container:

        >>> view.listIdsForDeletion()
        [u'pete']

    We must pass ids of selected people in the request:

        >>> request.form = {'delete.pete': 'on',
        ...                 'delete.john': 'on',
        ...                 'CONFIRM': 'Confirm'}
        >>> ids = [key for key in view.listIdsForDeletion()]
        >>> ids.sort()
        >>> ids
        [u'john', u'pete']
        >>> [item.title for item in view.itemsToDelete]
        ['Long John', 'Pete Parrot']

    These two should be gone after update:

        >>> view.update()
        Setting up table formatter:
            {'columns_before': [<...DependableCheckboxColumn ...>],
             'formatters': [<function url_cell_formatter ...>]}
        >>> ids = [key for key in personContainer]
        >>> ids.sort()
        >>> ids
        [u'frog', u'toad']

    If we press Cancel no one should get hurt though:

        >>> request.form = {'delete.frog': 'on',
        ...                 'delete.toad': 'on',
        ...                 'CANCEL': 'Cancel'}

    You see, both our firends are still in there:

        >>> ids = [key for key in personContainer]
        >>> ids.sort()
        >>> ids
        [u'frog', u'toad']

    """


def test_suite():
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF
                   | doctest.REPORT_ONLY_FIRST_FAILURE
                   | doctest.NORMALIZE_WHITESPACE)
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                       optionflags=optionflags))
    return suite


if __name__ == "__main__":
    unittest.main(defaultTest='test_suite')



#
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Schooltool grade levels.
"""

from persistent import Persistent

import zope.schema
from zope.interface import implements, implementer
from zope.component import adapter
from zope.container.btree import BTreeContainer
from zope.container.ordered import OrderedContainer
from zope.container.contained import Contained
from zope.annotation.interfaces import IAttributeAnnotatable

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.app import InitBase, StartUpBase
from schooltool.app.utils import TitledContainerItemVocabulary
from schooltool.relationship import URIObject, RelationshipSchema
from schooltool.relationship import RelationshipProperty
from schooltool.level import interfaces


URILevelCourses = URIObject(
    'http://schooltool.org/ns/cambodia/levelcourses/',
    'Courses at the level(s).',
    'Courses to be taught at given level(s).')

URICourse = URIObject(
    'http://schooltool.org/ns/cambodia/levelcourses/course',
    'Course',
    'The course to teach.')

URILevel = URIObject(
    'http://schooltool.org/ns/cambodia/levelcourses/level',
    'Level',
    'The teaching level.')


LevelCourses = RelationshipSchema(URILevelCourses,
                                 level=URILevel,
                                 course=URICourse)


LEVELS_APP_KEY = 'schooltool.level.level'


# BBB: for evolution
class LevelContainerContainer(BTreeContainer):
    """Container of level containers."""


class LevelContainer(OrderedContainer):
    """Container of levels."""

    implements(interfaces.ILevelContainer, IAttributeAnnotatable)


class Level(Persistent, Contained):
    """A grade level."""

    implements(interfaces.ILevel, interfaces.ILevelContained,
               IAttributeAnnotatable)
    courses = RelationshipProperty(URILevelCourses,
                                   URILevel,
                                   URICourse)

    def __init__(self, title=None):
        self.title = title


class VivifyLevelContainer(object):

    def __call__(self):
        if LEVELS_APP_KEY not in self.app:
            self.app[LEVELS_APP_KEY] = LevelContainer()


class LevelInit(VivifyLevelContainer, InitBase):
    pass


class LevelStartUp(VivifyLevelContainer, StartUpBase):
    pass


@adapter(ISchoolToolApplication)
@implementer(interfaces.ILevelContainer)
def getLevelContainer(app):
    return app[LEVELS_APP_KEY]


class LevelVocabulary(TitledContainerItemVocabulary):
    """Vocabulary of levels."""
    implements(zope.schema.interfaces.IIterableVocabulary)

    @property
    def container(self):
        app = ISchoolToolApplication(None)
        return interfaces.ILevelContainer(app, {})


def levelVocabularyFactory():
    return LevelVocabulary



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Notes support interfaces
"""
import zope.interface
import zope.schema

from zope.annotation.interfaces import IAnnotatable
from schooltool.common import SchoolToolMessage as _


class IHaveNotes(IAnnotatable):
    """An object that can have Notes.

    See also INote and INotes.
    """


class INote(zope.interface.Interface):
    """A note."""

    title = zope.schema.TextLine(
        title=_("Title"),
        description=_("Title of the note."))

    body = zope.schema.Text(
        title=_("Body"),
        description=_("Body of the note."))

    privacy = zope.schema.Choice(
        title=_("Privacy"),
        values=('private', 'public'),
        description=u"""
        Determines who can view the note.

        Can be one of two values: 'private', 'public'

        'private'  the note can only be viewed by the creator of the note.

        'public'   anyone can view the note, including anonymous users.

        """)

    owner = zope.interface.Attribute("""Component that owns this note.""")

    unique_id = zope.schema.TextLine(
        title=u"UID",
        required=False,
        description=u"""A globally unique id for this note.""")


class INotes(zope.interface.Interface):
    """A set of notes.

    Objects that can have notes are those that have an adapter to INotes.

    See also `INote`.
    """

    def __iter__():
        """Iterate over all notes."""

    def add(note):
        """Add a new note."""

    def remove(note):
        """Remove a note.

        Raises ValueError if note is not in the set.
        """

    def clear():
        """Remove all notes."""



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.generations.evolve33
"""
import unittest
import doctest

from zope.app.testing import setup

from zope.component import provideUtility
from zope.component import provideAdapter
from zope.interface import implements
from zope.intid import IntIds
from zope.intid.interfaces import IIntIds
from zope.site.folder import Folder
from zope.container.btree import BTreeContainer
from zope.component.hooks import getSite, setSite

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.generations.tests import ContextStub
from schooltool.contact.basicperson import getBoundContact
from schooltool.contact.interfaces import IContact
from schooltool.contact.interfaces import IUniqueFormKey
from schooltool.basicperson.interfaces import IBasicPerson
from schooltool.basicperson.person import BasicPerson


class AppStub(Folder):
    implements(ISchoolToolApplication)

    def __init__(self):
        super(AppStub, self).__init__()
        self['persons'] = BTreeContainer()
        self['persons']['john'] = BasicPerson("john", "Johny", "John")
        self['persons']['pete'] = BasicPerson("pete", "Petey", "Pete")


def verboseGetBoundContact(context):
    print 'IContact(<%s>)' % context.first_name
    return getBoundContact(context)



def doctest_evolve33():
    """Test evolution to generation 33.

    We'll need int ids.

        >>> provideUtility(IntIds(), IIntIds)

    Also an adapter to obtain the contact, and adapter to create form keys.

        >>> provideAdapter(verboseGetBoundContact, [IBasicPerson], IContact)
        >>> provideAdapter(lambda obj: obj.__name__, [None], IUniqueFormKey)

    And of course the app.

        >>> context = ContextStub()
        >>> context.root_folder['app'] = app = AppStub()
        >>> manager = setup.createSiteManager(app)

    Let's evolve now.  Note that persons get adapted to IContact.

        >>> from schooltool.generations.evolve33 import evolve
        >>> evolve(context)
        IContact(<Johny>)
        IContact(<Petey>)

    Site was restored after evolution.

        >>> print getSite()
        None

    """


from schooltool.generations.tests import catalogSetUp, catalogTearDown


def setUp(test):
    catalogSetUp(test)
    setup.setUpDependable()
    setSite()

def tearDown(test):
    setSite()
    catalogTearDown(test)


def test_suite():
    optionflags = (doctest.ELLIPSIS |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_ONLY_FIRST_FAILURE)
    return doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                optionflags=optionflags)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Timetables are meeting schedules created from scheduled day templates.
"""
import pytz
import datetime
import urllib

from persistent import Persistent
from persistent.list import PersistentList
from persistent.dict import PersistentDict
from zope.interface import implements
from zope.container.btree import BTreeContainer
from zope.component import getUtility
from zope.intid.interfaces import IIntIds

from schooltool.common import DateRange
from schooltool.timetable import interfaces
from schooltool.timetable.schedule import Meeting, Schedule
from schooltool.timetable.schedule import iterMeetingsInTimezone
from schooltool.timetable.schedule import iterMeetingsWithExceptions


class Timetable(Persistent, Schedule):
    implements(interfaces.ITimetable)

    periods = None
    time_slots = None
    exceptions = None

    def __init__(self, *args, **kw):
        Persistent.__init__(self)
        Schedule.__init__(self, *args, **kw)
        self.exceptions = PersistentDict()

    def periodMeetingId(self, date, period, meeting_n):
        date_id = date.isoformat()
        title = unicode(meeting_n)
        if (period is not None and
            period.title and
            period.title.strip()):
            title = unicode(period.title.strip())
            title = urllib.quote(title.encode('punycode'))
        uid = '%s.%s' % (date_id, title)
        return uid

    def iterOriginalMeetings(self, from_date, until_date=None):
        if until_date is None:
            until_date = from_date
        timezone = pytz.timezone(self.timezone)

        dates = DateRange(from_date, until_date)
        days = zip(dates,
                   self.periods.iterDates(dates),
                   self.time_slots.iterDates(dates))
        for day_date, day_periods, day_time_slots in days:
            if not day_periods:
                continue
            day = combineTemplates(day_periods, day_time_slots)
            for n, (period, time_slot) in enumerate(day):
                tstart = time_slot.tstart
                dtstart = datetime.datetime.combine(day_date, tstart)
                dtstart = timezone.localize(dtstart)
                # Note we're using dates in timetable's timezone here
                meeting_id = self.periodMeetingId(day_date, period, n+1)
                meeting = Meeting(
                    dtstart, time_slot.duration,
                    period=period,
                    meeting_id=meeting_id)
                yield meeting

    def iterMeetings(self, date, until_date=None):
        meetings = self.iterOriginalMeetings(date, until_date=until_date)

        return iterMeetingsWithExceptions(
            meetings, self.exceptions, self.timezone,
            date, until_date=until_date)


def combineTemplates(periods_template, time_slots_template):
    """Return ordered list of period, time_slot tuples."""
    result = []
    periods_by_activity = {}
    period_queue = periods_template.values()
    for period in period_queue:
        if period.activity_type not in periods_by_activity:
            periods_by_activity[period.activity_type] = []
        periods_by_activity[period.activity_type].append(period)

    for time_slot in sorted(time_slots_template.values()):
        activity = time_slot.activity_type
        if activity is not None:
            if (activity not in periods_by_activity or
                not periods_by_activity[activity]):
                continue # no more periods for this activity
            period = periods_by_activity[activity].pop(0)
            period_queue.remove(period)
            result.append((period, time_slot))
        else:
            period = period_queue.pop(0)
            periods_by_activity[period.activity_type].remove(period)
            result.append((period, time_slot))
    return result


class TimetableContainer(BTreeContainer):
    implements(interfaces.ITimetableContainer)

    default_id = None

    @property
    def default(self):
        if self.default_id is None:
            return None
        return self.get(self.default_id, None)

    @default.setter
    def default(self, timetable):
        if timetable is None:
            self.default_id = None
            return
        key = timetable.__name__
        if key not in self:
            raise KeyError(key)
        self.default_id = key


class SelectedPeriodsSchedule(Persistent, Schedule):
    implements(interfaces.ISelectedPeriodsSchedule)

    # XXX: think about storing intid here
    # or maybe better - a relationship
    timetable = None

    _periods = None

    consecutive_periods_as_one = False

    def __init__(self, timetable, *args, **kw):
        Schedule.__init__(self, *args, **kw)
        self.timetable = timetable
        self._periods = PersistentList()

    @property
    def periods(self):
        result = []
        for day in self.timetable.periods.templates.values():
            result.extend([period for period in day.values()
                           if self.hasPeriod(period)])
        return result

    def periodKey(self, period):
        day = period.__parent__
        return (day.__name__, period.__name__)

    def hasPeriod(self, period):
        key = self.periodKey(period)
        return key in self._periods

    def addPeriod(self, period):
        key = self.periodKey(period)
        if key not in self._periods:
            self._periods.append(key)

    def removePeriod(self, period):
        key = self.periodKey(period)
        if key in self._periods:
            self._periods.remove(key)

    def iterMeetings(self, date, until_date=None):
        if self.timetable is None:
            return
        meetings = iterMeetingsInTimezone(
            self.timetable, self.timezone, date, until_date=until_date)
        selected_periods = list(self.periods)

        last_meeting = None
        for orig_meeting in meetings:
            # XXX: proxy issues may breed here
            if orig_meeting.period in selected_periods:
                meeting = orig_meeting

                if (self.consecutive_periods_as_one and
                    last_meeting is not None and
                    meeting.dtstart.date() == last_meeting.dtstart.date()):
                    period_ids = meeting.period.__parent__.keys()
                    idx = period_ids.index(meeting.period.__name__)
                    if (idx > 0 and
                        period_ids[idx-1] == last_meeting.period.__name__):
                        meeting = meeting.clone(
                            meeting_id=last_meeting.meeting_id)

                last_meeting = orig_meeting

                yield meeting



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for registered security descriptions.
"""

import unittest
import doctest

from schooltool.schoolyear.testing import (setUp, tearDown,
                                           provideStubUtility,
                                           provideStubAdapter)
from schooltool.securitypolicy.ftesting import securitypolicy_functional_layer
from schooltool.securitypolicy.metaconfigure import getCrowdsUtility
from schooltool.securitypolicy.testing import (
    discriminator_sort_key,
    collectActionsByDiscriminator,
    printActionDescriptions,
    printDiscriminators)


def doctest_described_interfaces():
    r"""This is a crude attempt to list interfaces that are described in the
    security overview pages.

    First, collect interface/permission pairs with registered descriptions.

        >>> actions = collectActionsByDiscriminator()

        >>> sort_key = lambda i: discriminator_sort_key(i[0])
        >>> for discriminator, described in sorted(actions.items(),
        ...                                        key=sort_key):
        ...     print '%s: %s' % (
        ...         discriminator, [action.title for action in described])
        ('schooltool.edit', <...ISchoolToolApplication>):
            [u'Activate current school year',
        ...
             u'Modify']
        ('schooltool.view', <...ISchoolToolApplication>):
            [u'List', u'Access', u'View']
        ...
        ('schooltool.edit', <...IFieldDescription>):
            [u'Edit fields']
        ('schooltool.edit', <...ISchoolToolCalendar>):
            [u'Change calendar',
        ...

    Now, get the known crowds.

        >>> crowds = getCrowdsUtility().crowds.keys()

    And build sets of described and not described permission/interface pairs.

        >>> described = set(actions) & set(crowds)
        >>> missing = set(crowds) - set(actions)

    An interface/permission pair represents things the user can do in SchoolTool.

    Each interface/permission pair has a list of object/action pairs.
    The "object" is a rough approximation on SchoolTool object model from
    the user's perspective.  The "action" is an approximation of things user
    can do on the "object" if he has the required permission.

    Note the descriptions are sorted by modules of interfaces.  Some of the
    descriptions are declared in zcml, residing in other modules than the
    original interface.

    When you notice an "object" seemingly not belonging to the module of the
    interface (say, "Demographics"/"stuff" in schooltool.app.interfaces),
    this usually means one of the following:

    - The "object" has a functional extension on the described interface.

    - The "object" did not define it's own permissions and in reality inherits
      them from the parent.

    - There is a bug in zcml that registers the description.

        >>> described_actions = dict([
        ...     (discriminator, actions[discriminator])
        ...     for discriminator in described])

        >>> printActionDescriptions(described_actions)
        =========================
        schooltool.app.interfaces
        =========================
        - ISchoolToolApplication, schooltool.edit
        - ---------------------------------------
        -  School Years / Activate current school year
        -  Levels / Add/Remove
        -  School Years / Create/Delete
        -  Demographics / Manage
        -  SchoolTool application / Manage school settings and configuration
        -  Levels / Modify/Rename
        -  School Years / Modify
        -
        - ISchoolToolApplication, schooltool.view
        - ---------------------------------------
        -  School Years / List
        -  SchoolTool application / Access
        -  School Years / View
        -
        =================================
        schooltool.basicperson.interfaces
        =================================
        - IBasicPerson, schooltool.view
        - -----------------------------
        -  Messages / Browse other users' messages (overview)
        -
        - IFieldDescription, schooltool.edit
        - ----------------------------------
        -  Demographics / Edit fields
        -
        ==============================
        schooltool.calendar.interfaces
        ==============================
        - ISchoolToolCalendar, schooltool.edit
        - ------------------------------------
        -  SchoolTool application / Change calendar
        -  Sections / Change calendar
        -  Groups / Change calendar
        -  Users / Change calendar
        -  Reservations / Schedule reservation via calendar
        -
        - ISchoolToolCalendar, schooltool.view
        - ------------------------------------
        -  SchoolTool application / View calendar
        -  Sections / View calendar
        -  Groups / View calendar
        -  Users / View calendar
        -  Reservations / View reservation calendar
        -
        ==============================
        schooltool.contact.basicperson
        ==============================
        - IBoundContact, schooltool.view
        - ------------------------------
        -  Contacts / View user's contact information
        -
        =============================
        schooltool.contact.interfaces
        =============================
        - IContact, schooltool.edit
        - -------------------------
        -  Contacts / Modify an external contact
        -
        - IContact, schooltool.view
        - -------------------------
        -  Contacts / View an external contact
        -
        - IContactContainer, schooltool.edit
        - ----------------------------------
        -  Contacts / Create/Delete an external contact
        -
        - IContactContainer, schooltool.view
        - ----------------------------------
        -  Contacts / List/Search
        -
        ============================
        schooltool.course.interfaces
        ============================
        - ICourse, schooltool.edit
        - ------------------------
        -  Courses / Modify
        -
        - ICourse, schooltool.view
        - ------------------------
        -  Courses / View
        -
        - ICourseContainer, schooltool.edit
        - ---------------------------------
        -  Courses / Create/Delete
        -
        - ICourseContainer, schooltool.view
        - ---------------------------------
        -  Courses / List
        -
        - ISection, schooltool.edit
        - -------------------------
        -  Sections / Assign timetables
        -  Sections / Modify
        -
        - ISection, schooltool.view
        - -------------------------
        -  Sections / View
        -
        - ISectionContainer, schooltool.edit
        - ----------------------------------
        -  Sections / Create/Delete
        -
        - ISectionContainer, schooltool.view
        - ----------------------------------
        -  Sections / List
        -
        ===========================
        schooltool.group.interfaces
        ===========================
        - IGroup, schooltool.edit
        - -----------------------
        -  Groups / Modify
        -
        - IGroup, schooltool.view
        - -----------------------
        -  Groups / View
        -
        - IGroupContainer, schooltool.edit
        - --------------------------------
        -  Groups / Create/Delete
        -
        - IGroupContainer, schooltool.view
        - --------------------------------
        -  Groups / List
        -
        ============================
        schooltool.person.interfaces
        ============================
        - IPasswordWriter, schooltool.edit
        - --------------------------------
        -  Users / Change password
        -
        - IPerson, schooltool.edit
        - ------------------------
        -  Demographics / Modify user demographics
        -  Contacts / Manage user's contacts
        -  Contacts / Modify user's contact information
        -  Users / Modify
        -
        - IPerson, schooltool.editCalendarOverlays
        - ----------------------------------------
        -  Users / Manage visible calendars
        -
        - IPerson, schooltool.view
        - ------------------------
        -  Demographics / View user demographics
        -  Users / View
        -
        - IPersonContainer, schooltool.edit
        - ---------------------------------
        -  Users / Create/Delete
        -
        - IPersonContainer, schooltool.view
        - ---------------------------------
        -  Users / List/Search
        -
        - IPersonPreferences, schooltool.edit
        - -----------------------------------
        -  Users / Change preferences
        -
        - IPersonPreferences, schooltool.view
        - -----------------------------------
        -  Users / View preferences
        -
        ============================
        schooltool.report.interfaces
        ============================
        - IReportMessage, schooltool.view
        - -------------------------------
        -  Messages / Download reports
        -
        ==============================
        schooltool.resource.interfaces
        ==============================
        - IBaseResource, schooltool.edit
        - ------------------------------
        -  Reservations / Modify a resource
        -
        - IBaseResource, schooltool.view
        - ------------------------------
        -  Reservations / View a resource
        -
        - IResourceContainer, schooltool.edit
        - -----------------------------------
        -  Reservations / Add/Remove resources
        -
        - IResourceContainer, schooltool.view
        - -----------------------------------
        -  Reservations / List/Search resources
        -
        ==========================
        schooltool.task.interfaces
        ==========================
        - IMessage, schooltool.view
        - -------------------------
        -  Messages / Read messages
        -
        ==========================
        schooltool.term.interfaces
        ==========================
        - ITermContainer, schooltool.edit
        - -------------------------------
        -  Terms / Create/Delete
        -  Terms / Modify
        -
        - ITermContainer, schooltool.view
        - -------------------------------
        -  Terms / List
        -  Terms / View
        -
        ===============================
        schooltool.timetable.interfaces
        ===============================
        - IScheduleContainer, schooltool.edit
        - -----------------------------------
        -  Sections / Change schedule
        -
        - ITimetableContainer, schooltool.edit
        - ------------------------------------
        -  School timetables / Create/Delete
        -  School timetables / Modify
        -
        - ITimetableContainer, schooltool.view
        - ------------------------------------
        -  School timetables / List/Search
        -  School timetables / View
        -


    Some of the interface/permission pairs are intentionally (or by mistake!)
    not presented to the user.

        >>> printDiscriminators(missing)
        None, zope.ManageApplication
        None, zope.ManageContent
        None, zope.ManageServices
        None, zope.View
        None, zope.dublincore.change
        None, zope.dublincore.view
        -------------------------
        schooltool.app.interfaces
        -------------------------
        ISchoolToolApplication, zope.ManageSite
        ----------------------------
        schooltool.course.interfaces
        ----------------------------
        ICourseContainerContainer, schooltool.edit
        ----------------------------------
        schooltool.relationship.interfaces
        ----------------------------------
        IRelationshipLink, schooltool.edit
        IRelationshipLink, schooltool.view
        ----------------------------
        schooltool.report.interfaces
        ----------------------------
        IReportTask, schooltool.edit
        IReportTask, schooltool.view
        --------------------------
        schooltool.task.interfaces
        --------------------------
        IRemoteTask, schooltool.edit
        IRemoteTask, schooltool.view
        -------------------------------
        schooltool.timetable.interfaces
        -------------------------------
        IScheduleContainer, schooltool.view

        >>> print 'Total undescribed interface permissions: %d of %d (%d done)' % (
        ...     len(missing), len(crowds), len(actions))
        Total undescribed interface permissions: 15 of 57 (42 done)

    """


def test_suite():
    optionflags = (doctest.NORMALIZE_WHITESPACE |
                   doctest.ELLIPSIS |
                   doctest.REPORT_NDIFF)
    suite = doctest.DocTestSuite(optionflags=optionflags,
                                 extraglobs={'provideAdapter': provideStubAdapter,
                                             'provideUtility': provideStubUtility},
                                 setUp=setUp, tearDown=tearDown)
    suite.layer = securitypolicy_functional_layer
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



from zope.interface import Interface
from zope import schema
from zope.configuration import fields


class IEcho(Interface):

    message = schema.Text(title=u"Message", required=False)

    echo_on_add = fields.Bool(
        title=u"Echo when parsing ZCML.", required=False,
        default=False)


def handle_echo(message):
    print 'Executing echo:', message


def echo(_context, message=None, echo_on_add=False):
    discriminator=('echo', message)
    if echo_on_add:
        print 'Adding ZCML action: %s' % repr(discriminator)
    _context.action(discriminator=discriminator,
                    callable=handle_echo,
                    args=(message,))




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Macros
"""
import zope.interface

from zope.component import getMultiAdapter
from zope.publisher.browser import BrowserView

from zope.app import basicskin

class StandardMacros(basicskin.standardmacros.StandardMacros):
    macro_pages = ('view_macros', 'dialog_macros', 'rml_macros', )


class SchoolToolMacros(BrowserView):
    zope.interface.implements(zope.interface.common.mapping.IItemMapping)

    macro_pages = ('calendar_macros', 'generic_macros')

    def __getitem__(self, key):
        name = key + '_macros'

        if name in self.macro_pages:
            return getMultiAdapter((self.context, self.request), name=name)

        raise KeyError, key



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool XLS export views.
"""
import xlwt
import datetime
from operator import attrgetter
from StringIO import StringIO

from zope.interface import implements
from zope.security.proxy import removeSecurityProxy
from zope.publisher.browser import BrowserView
from zope.cachedescriptors.property import Lazy

from schooltool.skin import flourish
from schooltool.basicperson.demographics import DateFieldDescription
from schooltool.basicperson.interfaces import IDemographics, IBasicPerson
from schooltool.basicperson.interfaces import IDemographicsFields
from schooltool.common import SchoolToolMessage as _
from schooltool.group.interfaces import IGroupContainer
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import IAsset
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.common import format_time_range
from schooltool.contact.contact import URIPerson, URIContact
from schooltool.contact.contact import URIContactRelationship
from schooltool.contact.interfaces import IContact, IContactContainer
from schooltool.contact.interfaces import IContactable
from schooltool.export.interfaces import IXLSExportView
from schooltool.export.interfaces import IXLSProgressMessage
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.term.interfaces import ITermContainer
from schooltool.course.interfaces import ICourseContainer
from schooltool.course.interfaces import ISectionContainer
from schooltool.report.browser.report import RequestRemoteReportDialog
from schooltool.report.browser.report import ProgressReportPage
from schooltool.report.report import ReportLinkViewlet
from schooltool.relationship.relationship import IRelationshipLinks
from schooltool.task.progress import TaskProgress
from schooltool.task.progress import normalized_progress
from schooltool.report.report import AbstractReportTask
from schooltool.report.report import NoReportException
from schooltool.report.report import ReportFile
from schooltool.report.report import ReportMessage
from schooltool.report.report import OnPDFReportScheduled
from schooltool.timetable.interfaces import ITimetableContainer
from schooltool.timetable.interfaces import IScheduleContainer
from schooltool.timetable.interfaces import IHaveTimetables
from schooltool.timetable.daytemplates import CalendarDayTemplates
from schooltool.timetable.daytemplates import WeekDayTemplates
from schooltool.timetable.daytemplates import SchoolDayTemplates


class ExcelExportView(ProgressReportPage):

    implements(IXLSExportView)

    def __init__(self, context, request):
        super(ExcelExportView, self).__init__(context, request)
        self._font_cache = {}
        self._style_cache = {}

    def setUpHeaders(self, data):
        """Set up HTTP headers to serve data as excel spreadsheet."""
        response = self.request.response
        response.setHeader('Content-Type', 'application/vnd.ms-excel')
        response.setHeader('Content-Length', len(data))

    def listIds(self, header, items, ws, offset, last=False):
        if not items:
            return offset - 1
        self.write_header(ws, offset + 1, 0,  header, merge=1)
        for n, item in enumerate(sorted(items, key=lambda i: i.__name__)):
            self.write(ws, offset + 2 + n, 0,  item.__name__)
            self.write(ws, offset + 2 + n, 1,  "")
        return 1 + offset + len(items)

    def skipRow(self, ws, offset):
        return offset + 1

    def listFields(self, item, accessors, ws, offset):
        for n, accessor in enumerate(accessors):
            header, field, style = accessor(item)
            self.write_header(ws, offset + n, 0, header)
            self.write(ws, offset + n, 1, field, format_str=style)
        return offset + len(accessors)

    def _makeFont(self, font_key):
        font = xlwt.Font()
        for attr, value in font_key:
            setattr(font, attr, value)
        return font

    def getFont(self, font_key):
        font = self._font_cache.get(font_key, None)
        if font is None:
            self._font_cache[font_key] = font = self._makeFont(font_key)
        return font

    def write(self, ws, row, col, data,
              bold=False,
              color=None,
              format_str=None,
              borders=None,
              merge=None):
        if borders is None:
            borders = []
        if data is None:
            data = ""
        if type(data) == type(True):
            data = str(data)
        key = (bold, color, format_str, tuple(borders))
        style = self._style_cache.get(key, None)
        if style is None:
            style = xlwt.XFStyle()
            if bold:
                style.font = self.getFont((("bold", True),))

            if color is not None:
                pattern = xlwt.Pattern()
                pattern.pattern = xlwt.Pattern.SOLID_PATTERN
                pattern.pattern_fore_colour = color
                style.pattern = pattern

            if format_str is not None:
                style.num_format_str = format_str

            if borders:
                b = xlwt.Formatting.Borders()
                for border in borders:
                    setattr(b, border, xlwt.Formatting.Borders.THIN)
                style.borders = b
            self._style_cache[key] = style
        if merge is None:
            ws.write(row, col, data, style)
        else:
            ws.write_merge(row, row, col, col + merge, data, style)

    def write_header(self, ws, row, col, data,
                     borders=None, merge=None):
        YELLOW = 5
        self.write(ws, row, col, data, borders=borders, bold=True, color=YELLOW,
                   merge=merge)


class SchoolTimetableExportView(ExcelExportView):

    dows = ['Monday', 'Tuesday', 'Wednesday', 'Thursday',
            'Friday', 'Saturday', 'Sunday']

    def format_periods(self, periods, ws, offset):
        self.write_header(ws, offset, 0, "Days")
        max_periods = max([len(day) for day in periods.templates.values()])
        self.write_header(ws, offset, 1, "Periods", merge=max_periods - 1)
        offset += 1

        for day in periods.templates.values():
            self.write(ws, offset, 0, day.title)
            for col, period in enumerate(day.values()):
                self.write(ws, offset, col + 1, period.title)
                self.write(ws, offset + 1, col + 1, period.activity_type)
            offset += 2
        return offset

    def format_time_slots(self, time_slots, ws, offset):
        self.write_header(ws, offset, 0, "Time schedule")
        offset += 1

        for day in time_slots.templates.values():
            self.write(ws, offset, 0, day.title)
            for col, slot in enumerate(day.values()):
                time = format_time_range(slot.tstart, slot.duration)
                self.write(ws, offset, col + 1, time)
                self.write(ws, offset + 1, col + 1, slot.activity_type)
            offset += 2
        return offset

    day_templates = (
        ('calendar_days', CalendarDayTemplates),
        ('week_days', WeekDayTemplates),
        ('school_days', SchoolDayTemplates),
        )

    def format_school_timetable(self, timetable, ws, offset):
        template_ids = dict([(cls, tid)
                             for tid, cls in self.day_templates])
        factory_id = lambda t: template_ids[t.__class__]
        schoolyear_id = lambda t: IHaveTimetables(t).__name__
        fields = [lambda i: ("School Timetable", i.title, None),
                  lambda i: ("ID", i.__name__, None),
                  lambda i: ("School Year", schoolyear_id(i), None),
                  lambda i: ("Period days", factory_id(i.periods), None),
                  lambda i: ("Time slots", factory_id(i.time_slots), None)]

        offset = self.listFields(timetable, fields, ws, offset)
        offset = self.skipRow(ws, offset)
        offset = self.format_periods(timetable.periods, ws, offset)
        offset = self.skipRow(ws, offset)
        offset = self.format_time_slots(timetable.time_slots, ws, offset)
        return offset + 1

    def export_school_timetables(self, wb):
        self.task_progress.force('export_school_timetables', active=True)
        ws = wb.add_sheet("School Timetables")
        school_years = sorted(ISchoolYearContainer(self.context).values(),
                              key=lambda s: s.first)
        row = 0
        for ny, school_year in enumerate(sorted(school_years, key=lambda i: i.last)):
            timetables = ITimetableContainer(school_year)
            for nt, timetable in enumerate(sorted(timetables.values(), key=lambda i: i.__name__)):
                row = self.format_school_timetable(timetable, ws, row) + 1
                self.progress('export_school_timetables', normalized_progress(
                        ny, len(school_years), nt, len(timetables)))
        self.finish('export_school_timetables')


def merge_date_ranges(dates):
    ranges = []
    start = None
    previous_day = None
    for day in dates:
        if start is None:
            previous_day = start = day
        elif day - datetime.timedelta(days=1) == previous_day:
            previous_day = day
        else:
            ranges.append((start, previous_day))
            previous_day = start = day
    if start is not None:
        ranges.append((start, previous_day))

    return ranges


class Text(object):

    style = {}
    def __init__(self, data):
        self.data = data

    def __cmp__(self, other):
        return cmp(self.data, other.data)

    def __repr__(self):
        return 'Text(%r)' % self.data


class Header(Text):
    YELLOW = 5
    style = {'bold': True,
             'color': YELLOW}

    def __repr__(self):
        return 'Header(%r)' % self.data


class Date(Text):
    style = {'format_str': 'YYYY-MM-DD'}

    def __repr__(self):
        return 'Date(%r)' % self.data


class ContactRelationship(object):

    def __init__(self, person, contact, relationship):
        self.person = person
        self.contact = contact
        self.relationship = relationship


class MegaExporter(SchoolTimetableExportView):

    overall_line_id = 'overall'

    def print_table(self, table, ws):
        for x, row in enumerate(table):
            for y, cell in enumerate(row):
                self.write(ws, x, y, cell.data, **cell.style)
        return len(table)

    def format_table(self, fields, items, importer=None, major_progress=()):
        headers = [Header(header)
                   for header, style, getter in fields]
        rows = []
        total_items = len(removeSecurityProxy(items))
        for n, item in enumerate(items):
            row = [style(getter(item))
                   for header, style, getter in fields]
            rows.append(row)
            if importer is not None:
                self.progress(importer, normalized_progress(
                        *(major_progress + (n, total_items))))
        rows.sort()
        return [headers] + rows

    def format_school_years(self):
        fields = [('ID', Text, attrgetter('__name__')),
                  ('Title', Text, attrgetter('title')),
                  ('Start', Date, attrgetter('first')),
                  ('End', Date, attrgetter('last'))]
        items = ISchoolYearContainer(self.context).values()
        result = self.format_table(fields, items, importer='export_school_years')
        return result

    def export_school_years(self, wb):
        self.task_progress.force('export_school_years', active=True)
        ws = wb.add_sheet("School Years")
        self.print_table(self.format_school_years(), ws)
        self.finish('export_school_years')

    def calculate_holidays_and_weekdays(self):

        work_days = 0.0

        days_of_week = {}
        for dow in range(7):
            days_of_week[dow] = [0, 0]

        school_years = ISchoolYearContainer(self.context).values()
        for school_year in school_years:
            terms = ITermContainer(school_year).values()
            for term in terms:
                for date in term:
                    if term.isSchoolday(date):
                        days_of_week[date.weekday()][0] += 1
                        work_days += 1
                    else:
                        days_of_week[date.weekday()][1] += 1

        if work_days == 0:
            return [[], list(range(7)), []]

        coefficients = [counts[0] / work_days
                        for day, counts in sorted(days_of_week.items())]

        # Weekends
        weekends = []
        for n, k in enumerate(coefficients):
            if k < 0.1:
                weekends.append(n)

        # Weekend exceptions and holidays
        holidays = []
        weekend_exceptions = []
        for school_year in school_years:
            terms = ITermContainer(school_year).values()
            for term in terms:
                for date in term:
                    if term.isSchoolday(date) and date.weekday() in weekends:
                        weekend_exceptions.append(date)
                    elif not term.isSchoolday(date) and date.weekday() not in weekends:
                        holidays.append(date)

        holiday_ranges = merge_date_ranges(holidays)
        return [holiday_ranges, weekends, weekend_exceptions]

    def format_holidays(self, holidays):
        if not holidays:
            return []
        table = [[], [Header("Holidays")]]
        table.extend([[Date(start), Date(end)]
                      for start, end in holidays])
        return table

    def format_weekends(self, weekends):
        if not weekends:
            return []

        table = [[], [Header("Weekends")]]
        weekdays = map(Text, ['Monday',
                              'Tuesday',
                              'Wednesday',
                              'Thursday',
                              'Friday',
                              'Saturday',
                              'Sunday'])
        table.append(weekdays)
        table.append([weekday in weekends and Text('X') or Text('')
                      for weekday in range(len(weekdays))])
        return table

    def format_weekend_exceptions(self, working_weekends):
        if not working_weekends:
            return []

        table = [[], [Header("Working weekends")]]
        table.extend([[Date(day)]
                      for day in working_weekends])
        return table

    def format_terms(self):
        fields = [('SchoolYear', Text, lambda t: t.__parent__.__name__),
                  ('ID', Text, attrgetter('__name__')),
                  ('Title', Text, attrgetter('title')),
                  ('Start', Date, attrgetter('first')),
                  ('End', Date, attrgetter('last'))]

        school_years = ISchoolYearContainer(self.context).values()
        items = []
        for year in school_years:
            items.extend([term for term in
                          ITermContainer(year).values()])
        terms_table = self.format_table(fields, items, importer='export_terms')
        holidays, weekends, exceptions = self.calculate_holidays_and_weekdays()
        terms_table.extend(self.format_holidays(holidays))
        terms_table.extend(self.format_weekends(weekends))
        terms_table.extend(self.format_weekend_exceptions(exceptions))
        return terms_table

    def export_terms(self, wb):
        self.task_progress.force('export_terms', active=True)
        ws = wb.add_sheet("Terms")
        self.print_table(self.format_terms(), ws)
        self.finish('export_terms')

    def format_persons(self):
        fields = [('User Name', Text, attrgetter('__name__')),
                  ('Prefix', Text, attrgetter('prefix')),
                  ('First Name', Text, attrgetter('first_name')),
                  ('Middle Name', Text, attrgetter('middle_name')),
                  ('Last Name', Text, attrgetter('last_name')),
                  ('Suffix', Text, attrgetter('suffix')),
                  ('Preferred Name', Text, attrgetter('preferred_name')),
                  ('Birth Date', Date, attrgetter('birth_date')),
                  ('Gender', Text, attrgetter('gender')),
                  ('Password', Text, lambda p: None)]

        def demographics_getter(attribute):
            def getter(person):
                demographics = IDemographics(person)
                return demographics[attribute]
            return getter

        app = ISchoolToolApplication(None)
        demographics_fields = IDemographicsFields(app)
        for field in demographics_fields.values():
            title = field.title
            format = Text
            if isinstance(field, DateFieldDescription):
                format = Date
            getter = demographics_getter(field.name)
            fields.append((title, format, getter))

        items = self.context['persons'].values()
        return self.format_table(fields, items, importer='export_persons')

    def export_persons(self, wb):
        self.task_progress.force('export_persons', active=True)
        ws = wb.add_sheet("Persons")
        self.print_table(self.format_persons(), ws)
        self.finish('export_persons')

    def format_contact_persons(self):
        def contact_getter(attribute):
            def getter(contact):
                person = IBasicPerson(contact.__parent__, None)
                if person is None:
                    return getattr(contact, attribute)
                if attribute == '__name__':
                    return person.username
                return ''
            return getter

        fields = [('ID', Text, contact_getter('__name__')),
                  ('Prefix', Text, contact_getter('prefix')),
                  ('First Name', Text, contact_getter('first_name')),
                  ('Middle Name', Text, contact_getter('middle_name')),
                  ('Last Name', Text, contact_getter('last_name')),
                  ('Suffix', Text, contact_getter('suffix')),
                  ('Address line 1', Text, attrgetter('address_line_1')),
                  ('Address line 2', Text, attrgetter('address_line_2')),
                  ('City', Date, attrgetter('city')),
                  ('State', Date, attrgetter('state')),
                  ('Country', Date, attrgetter('country')),
                  ('Postal code', Date, attrgetter('postal_code')),
                  ('Home phone', Text, attrgetter('home_phone')),
                  ('Work phone', Text, attrgetter('work_phone')),
                  ('Mobile phone', Text, attrgetter('mobile_phone')),
                  ('Email', Text, attrgetter('email')),
                  ('Language', Text, attrgetter('language'))]

        items = []
        for person in self.context['persons'].values():
            items.append(IContact(person))
        for contact in IContactContainer(self.context).values():
            items.append(contact)

        return self.format_table(fields, items, importer='export_contacts',
                                 major_progress=(0,2))

    def format_contact_relationships(self):
        fields = [('Person ID', Text, attrgetter('person')),
                  ('Contact ID', Text, attrgetter('contact')),
                  ('Relationship', Text, attrgetter('relationship'))]

        items = []
        for person in self.context['persons'].values():
            person = removeSecurityProxy(person)
            for contact in IContactable(person).contacts:
                try:
                    links = IRelationshipLinks(person)
                    link = links.find(
                        URIPerson, contact, URIContact, URIContactRelationship)
                except ValueError:
                    continue
                target_person = IBasicPerson(contact.__parent__, None)
                if target_person is None:
                    name = contact.__name__
                else:
                    name = target_person.username
                item = ContactRelationship(person.username,
                    name, link.extra_info.relationship)
                items.append(item)

        return self.format_table(fields, items, importer='export_contacts',
                                 major_progress=(1,2))

    def export_contacts(self, wb):
        self.task_progress.force('export_contacts', active=True)
        ws = wb.add_sheet("Contact Persons")
        self.print_table(self.format_contact_persons(), ws)
        ws = wb.add_sheet("Contact Relationships")
        self.print_table(self.format_contact_relationships(), ws)
        self.finish('export_contacts')

    def format_resources(self):
        fields = [('ID', Text, attrgetter('__name__')),
                  ('Type', Text, lambda r: r.__class__.__name__),
                  ('Title', Text, attrgetter('title'))]
        items = self.context['resources'].values()
        return self.format_table(fields, items, importer='export_resources')

    def export_resources(self, wb):
        self.task_progress.force('export_resources', active=True)
        ws = wb.add_sheet("Resources")
        self.print_table(self.format_resources(), ws)
        self.finish('export_resources')

    def format_courses(self):
        fields = [('School Year', Text, lambda c: ISchoolYear(c).__name__),
                  ('ID', Text, attrgetter('__name__')),
                  ('Title', Text, attrgetter('title')),
                  ('Description', Text, attrgetter('description')),
                  ('Local ID', Text, attrgetter('course_id')),
                  ('Government ID', Text, attrgetter('government_id')),
                  ('Credits', Text, attrgetter('credits'))]

        school_years = ISchoolYearContainer(self.context).values()
        items = []
        for year in school_years:
            items.extend([term for term in
                          ICourseContainer(year).values()])
        return self.format_table(fields, items, importer='export_courses')

    def export_courses(self, wb):
        self.task_progress.force('export_courses', active=True)
        ws = wb.add_sheet("Courses")
        self.print_table(self.format_courses(), ws)
        self.finish('export_courses')

    def format_timetables(self, section, ws, offset):
        schedules = IScheduleContainer(section)
        if not schedules:
            return offset
        schedules = list(sorted(schedules.values(),
                                key=lambda s: s.timetable.__name__))
        for schedule in schedules:
            timetable = schedule.timetable
            self.write_header(ws, offset, 0,  "School Timetable")
            self.write(ws, offset, 1,  timetable.__name__)
            offset += 1

            self.write(ws, offset, 0,  "Consecutive periods as one")
            self.write(ws, offset, 1,
                       schedule.consecutive_periods_as_one and 'yes' or 'no')
            offset += 1

            self.write_header(ws, offset, 0,  "Day")
            self.write_header(ws, offset, 1,  "Period")
            offset += 1

            for period in schedule.periods:
                day = period.__parent__
                self.write(ws, offset, 0,  day.title)
                self.write(ws, offset, 1,  period.title)
                offset += 1
            offset += 1
        return offset

    def format_section(self, year, courses, term, section, ws, row):
        teachers = [t.__name__ for t in section.instructors]
        resources = [r.__name__ for r in section.resources]
        self.write(ws, row, 0, year.__name__)
        self.write(ws, row, 1, courses)
        self.write(ws, row, 2, term.__name__)
        self.write(ws, row, 3, section.__name__)
        if section.previous:
            self.write(ws, row, 4, section.previous.__name__)
        if section.next:
            self.write(ws, row, 5, section.next.__name__)
        self.write(ws, row, 6, section.title)
        if section.description:
            self.write(ws, row, 7, section.description)
        self.write(ws, row, 8, ', '.join(teachers))
        self.write(ws, row, 9, ', '.join(resources))

    def export_sections(self, wb):
        self.task_progress.force('export_sections', active=True)
        ws = wb.add_sheet("Sections")
        headers = ["School Year", "Courses", "Term", "Section ID",
                   "Previous ID", "Next ID", "Title", "Description",
                   "Instructors", "Resources"]
        for index, header in enumerate(headers):
            self.write_header(ws, 0, index, header)

        sections = []
        for year in ISchoolYearContainer(self.context).values():
            for term in year.values():
                for section in ISectionContainer(term).values():
                    if not list(section.courses):
                        continue
                    courses = ', '.join([c.__name__ for c in section.courses])
                    sections.append((year, courses, term.first, term,
                                     section.__name__, section))

        row = 1
        sections.sort()
        n_sections = len(sections)
        for n, (year, courses, first, term, section_id, section) in enumerate(sections):
            self.format_section(year, courses, term, section, ws, row)
            self.progress('export_sections', normalized_progress(
                    n, n_sections))
            row += 1
        self.finish('export_sections')

    def format_student_sections(self, year, student_sections, ws, row):
        headers = ["School Year", "Term", "Section ID"]
        for index, header in enumerate(headers):
            self.write_header(ws, row, index, header)
        row += 1

        for first, term, section_id, section in sorted(student_sections):
            self.write(ws, row, 0, year.__name__)
            self.write(ws, row, 1, term.__name__)
            self.write(ws, row, 2, section.__name__)
            row += 1
        return row + 1

    def format_students_block(self, students, ws, row):
        self.write_header(ws, row, 0, "Students")
        row += 1

        for student in students.split(','):
            self.write(ws, row, 0, student)
            row += 1
        return row + 1

    def export_section_enrollment(self, wb):
        self.task_progress.force('export_section_enrollment', active=True)
        ws = wb.add_sheet("SectionEnrollment")

        year_sections = {}
        years = ISchoolYearContainer(self.context)
        for ny, year in enumerate(years.values()):
            sections = year_sections[year] = {}
            for nt, term in enumerate(year.values()):
                term_sections = ISectionContainer(term)
                for ns, section in enumerate(term_sections.values()):
                    if not list(section.courses):
                        continue
                    student_ids = [s.username for s in section.members]
                    students = ','.join(sorted(student_ids))
                    student_sections = sections.setdefault(students, [])
                    student_sections.append((term.first, term, section.__name__,
                                             section))
                    self.progress('export_section_enrollment', normalized_progress(
                        ny, len(years),
                        nt, len(year),
                        ns, len(term_sections),
                        ) * 0.9)

        row = 0
        for ny, (year, sections) in enumerate(sorted(year_sections.items())):
            for ns, (students, student_sections) in enumerate(sorted(sections.items())):
                row = self.format_student_sections(year, student_sections, ws,
                                                   row)
                row = self.format_students_block(students, ws, row)
                self.progress('export_section_enrollment', normalized_progress(
                        ny, len(year_sections),
                        ns, len(sections),
                        ) * 0.1 + 0.9)
        self.finish('export_section_enrollment')

    def format_timetable_sections(self, year, timetable_sections, ws, row):
        headers = ["School Year", "Term", "Section ID"]
        for index, header in enumerate(headers):
            self.write_header(ws, row, index, header)
        row += 1

        for first, term, section_id, section in sorted(timetable_sections):
            self.write(ws, row, 0, year.__name__)
            self.write(ws, row, 1, term.__name__)
            self.write(ws, row, 2, section.__name__)
            row += 1
        return row + 1

    def format_timetables_block(self, timetables, ws, row):
        for timetable in timetables:
            parts = timetable.split(',')
            self.write_header(ws, row, 0, "Timetable")
            self.write(ws, row, 1, parts[0])
            self.write_header(ws, row, 2, "Consecutive")
            self.write(ws, row, 3, parts[1])
            row += 2

            self.write_header(ws, row, 0, "Day")
            self.write_header(ws, row, 1, "Period ID")
            row += 1

            parts = parts[2:]
            while parts:
                day, period = parts[:2]
                self.write(ws, row, 0, day)
                self.write(ws, row, 1, period)
                parts = parts[2:]
                row += 1
        return row + 1

    def export_section_timetables(self, wb):
        self.task_progress.force('export_section_timetables', active=True)
        ws = wb.add_sheet("SectionTimetables")

        year_sections = {}
        for year in ISchoolYearContainer(self.context).values():
            sections = year_sections[year] = {}
            for term in year.values():
                for section in ISectionContainer(term).values():
                    if not list(section.courses):
                        continue
                    timetables = []
                    for schedule in IScheduleContainer(section).values():
                        if schedule.timetable.__name__ is None:
                            # LP: #1281335
                            continue
                        parts = [schedule.timetable.__name__]
                        if schedule.consecutive_periods_as_one:
                            parts.append('yes')
                        else:
                            parts.append('no')
                        for period in schedule.periods:
                            day = period.__parent__
                            parts.append(day.title)
                            parts.append(period.title)
                        timetables.append(','.join(parts))
                    if not len(timetables):
                        continue
                    timetables = tuple(timetables)
                    timetable_sections = sections.setdefault(timetables, [])
                    timetable_sections.append((term.first, term,
                                               section.__name__, section))

        row = 0
        for ny, (year, sections) in enumerate(sorted(year_sections.items())):
            for nt, (timetables, timetable_sections) in enumerate(sorted(sections.items())):
                row = self.format_timetable_sections(year, timetable_sections,
                                                     ws, row)
                row = self.format_timetables_block(timetables, ws, row)
                self.progress('export_section_timetables', normalized_progress(
                        ny, len(year_sections), nt, len(sections)
                        ))
        self.finish('export_section_timetables')

    def format_group(self, group, ws, offset):
        fields = [lambda i: ("Group Title", i.title, None),
                  lambda i: ("ID", i.__name__, None),
                  lambda i: ("School Year", ISchoolYear(i.__parent__).__name__, None),
                  lambda i: ("Description", i.description, None)]

        offset = self.listFields(group, fields, ws, offset)
        offset = self.listIds("Members", group.members, ws, offset) + 1
        offset = self.listIds("Leaders", IAsset(group).leaders, ws, offset,
                              last=True) + 1
        return offset

    def export_groups(self, wb):
        self.task_progress.force('export_groups', active=True)
        ws = wb.add_sheet("Groups")
        school_years = sorted(ISchoolYearContainer(self.context).values(),
                              key=lambda s: s.first)
        row = 0
        for ny, school_year in enumerate(sorted(school_years, key=lambda i: i.last)):
            groups = IGroupContainer(school_year)
            for ng, group in enumerate(sorted(groups.values(), key=lambda i: i.__name__)):
                row = self.format_group(group, ws, row) + 1
                self.progress('export_groups', normalized_progress(
                        ny, len(school_years), ng, len(groups)
                        ))
        self.finish('export_groups')

    def makeProgress(self):
        self.task_progress = TaskProgress(None)

    def addImporters(self, progress):
        progress.add('export_school_years', active=False,
                     title=_('School Years'), progress=0.0)
        progress.add('export_terms', active=False,
                     title=_('Terms'), progress=0.0)
        progress.add('export_school_timetables', active=False,
                     title=_('School Timetables'), progress=0.0)
        progress.add('export_resources', active=False,
                     title=_('Resources'), progress=0.0)
        progress.add('export_persons', active=False,
                     title=_('Persons'), progress=0.0)
        progress.add('export_contacts', active=False,
                     title=_('Contacts'), progress=0.0)
        progress.add('export_courses', active=False,
                     title=_('Courses'), progress=0.0)
        progress.add('export_sections', active=False,
                     title=_('Sections'), progress=0.0)
        progress.add('export_section_enrollment', active=False,
                     title=_('Section Enrollment'), progress=0.0)
        progress.add('export_section_timetables', active=False,
                     title=_('Section Schedules'), progress=0.0)
        progress.add('export_groups', active=False,
                     title=_('Groups'), progress=0.0)
        progress.add('overall',
                     title=_('School Data'), progress=0.0)

    def update(self):
        super(MegaExporter, self).update()
        self.addImporters(self.task_progress)

    def render(self, workbook):
        datafile = StringIO()
        workbook.save(datafile)
        data = datafile.getvalue()
        self.setUpHeaders(data)
        return data

    def __call__(self):
        self.makeProgress()
        self.task_progress.title = _("Exporting school data")
        self.addImporters(self.task_progress)

        wb = xlwt.Workbook()
        self.export_school_years(wb)
        self.export_terms(wb)
        self.export_school_timetables(wb)
        self.export_resources(wb)
        self.export_persons(wb)
        self.export_contacts(wb)
        self.export_courses(wb)
        self.export_sections(wb)
        self.export_section_enrollment(wb)
        self.export_section_timetables(wb)
        self.export_groups(wb)
        self.task_progress.title = _("Export complete")
        self.task_progress.force('overall', progress=1.0)
        data = self.render(wb)
        return data


class XLSReportTask(AbstractReportTask):

    default_filename = 'report.xls'
    default_mimetype = 'application/vnd.ms-excel'

    def renderReport(self, renderer, stream, *args, **kw):
        workbook = renderer(*args, **kw)
        if workbook is None:
            raise NoReportException()
        workbook.save(stream)


class RemoteMegaExporter(MegaExporter):

    base_filename = 'school'
    message_title = _('school export')
    makeProgress = ExcelExportView.makeProgress

    def render(self, workbook):
        # Return the workbook itself, should use it to save to blob directly
        return workbook


class RequestXLSReportDialog(RequestRemoteReportDialog):

    task_factory = XLSReportTask


class XLSProgressMessage(ReportMessage):
    implements(IXLSProgressMessage)


class OnXLSReportScheduled(OnPDFReportScheduled):

    message_factory = XLSProgressMessage

    def makeReportTitle(self):
        title = getattr(self.view, 'message_title', None)
        if not title:
            title = getattr(self.view, 'filename', None)
        if not title:
            title = _(u'XLS export')
        return title


class ExportLinkVielwet(ReportLinkViewlet):
    pass


class ImportLinkViewlet(flourish.page.LinkViewlet):
    pass



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Testing utilities for testing Term related code.
"""
import datetime
from persistent import Persistent

from zope.location.location import LocationProxy
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.interface import implementer
from zope.interface import implements
from zope.component import adapter
from zope.component import getUtility
from zope.component import provideUtility

from z3c.form import form, field

from schooltool.utility.utility import UtilitySetUp
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.term.term import getNextTermForDate
from schooltool.term.interfaces import IDateManager
from schooltool.common import SchoolToolMessage as _


class DateManagerStub(object):
    implements(IDateManager)

    def __init__(self, today, current_term):
        self.today = today
        self.current_term = current_term


def setUpDateManagerStub(today=None, current_term=None):
    if not today:
        today = datetime.date(2005, 9, 20)
    provideUtility(DateManagerStub(today, current_term))


class LocalDateManagerUtility(Persistent, DateManagerStub):
    implements(IDateManager)

    def __init__(self):
        self.today = datetime.date(2005, 2, 1)

    @property
    def current_term(self):
        return getNextTermForDate(self.today)


dateManagerSetupSubscriber = UtilitySetUp(
    LocalDateManagerUtility, IDateManager)


@implementer(IDateManager)
@adapter(ISchoolToolApplication)
def getDateManager(context):
    return LocationProxy(getUtility(IDateManager), context, 'time')


class DateManagementView(form.EditForm):
    label = _("Set the date for the school")
    template = ViewPageTemplateFile('date_management.pt')

    form.extends(form.EditForm)
    fields = field.Fields(IDateManager)

    def updateActions(self):
        super(DateManagementView, self).updateActions()
        self.actions['apply'].addClass('button-ok')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool flourish resource extension.
"""
import zope.schema

import zc.resourcelibrary
import zc.resourcelibrary.resourcelibrary
from zope.component import adapts, queryAdapter
from zope.interface import implements, Interface, Attribute
from zope.publisher.interfaces.browser import IBrowserRequest


class IResourceLibrary(Interface):
    """Dynamic resource library."""

    __name__ = zope.schema.TextLine(
        title=u"Library name")

    configure = Attribute("Configure the library class")

    required = zope.schema.List(
        title=u"Required libraries",
        value_type=zope.schema.TextLine())

    included = zope.schema.List(
        title=u"Included files in this library",
        value_type=zope.schema.TextLine())


_hooks = None


def getRequired(name):
    request = zc.resourcelibrary.resourcelibrary.getRequest()
    lib = queryAdapter(request, IResourceLibrary, name=name)
    if lib is None:
        global _hooks
        if _hooks is not None:
            return _hooks['getRequired'](name)
        return ()
    return lib.required


def getIncluded(name, alternative=None):
    request = zc.resourcelibrary.resourcelibrary.getRequest()
    lib = queryAdapter(request, IResourceLibrary, name=name)
    if lib is None:
        global _hooks
        if _hooks is not None:
            return _hooks['getIncluded'](name)
        return ()
    return lib.included


def patch_zc_resourcelibrary():
    global _hooks
    if _hooks is None:
        _hooks = {
            'getRequired': zc.resourcelibrary.getRequired,
            'getIncluded': zc.resourcelibrary.getIncluded,
            }
        # Monkeypatch zc.resourcelibrary
        zc.resourcelibrary.getRequired = getRequired
        zc.resourcelibrary.getIncluded = getIncluded


def unpatch_zc_resourcelibrary():
    global _hooks
    if _hooks is not None:
        zc.resourcelibrary.getRequired = _hooks['getRequired']
        zc.resourcelibrary.getIncluded = _hooks['getIncluded']
    _hooks = None


try:
    from zope.testing.cleanup import addCleanUp
except ImportError:
    pass
else:
    addCleanUp(unpatch_zc_resourcelibrary)
    del addCleanUp


class ResourceLibrary(object):
    adapts(IBrowserRequest)
    implements(IResourceLibrary)

    __name__ = None

    def __init__(self, request):
        self.request = request

    @classmethod
    def configure(cls):
        pass

    @property
    def required(self):
        global _hooks
        if _hooks is not None:
            return _hooks['getRequired'](self.__name__)
        return ()

    @property
    def included(self):
        global _hooks
        if _hooks is not None:
            return _hooks['getIncluded'](self.__name__)
        return ()



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Upgrade SchoolTool to generation 34.

Evolution script to set the 'enabled' attribute in existing EmailContainers.
"""
from zope.app.generations.utility import findObjectsProviding
from zope.app.publication.zopepublication import ZopePublication

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.email.interfaces import IEmailContainer


def evolve(context):
    root = context.connection.root().get(ZopePublication.root_name, None)

    apps = findObjectsProviding(root, ISchoolToolApplication)
    for app in apps:
        container = IEmailContainer(app, None)
        if container is None:
            continue
        if getattr(container, 'enabled', None) is None:
            container.enabled = bool(container.hostname)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool flourish skin interfaces.
"""

import zope.interface.interfaces
import zope.schema
import zope.schema.interfaces
import zope.viewlet.interfaces
import zope.contentprovider.interfaces
from zope.interface import Interface, Attribute
from zope.location.interfaces import ILocation
from zope.publisher.interfaces import IPublishTraverse
from zope.publisher.interfaces.browser import IDefaultBrowserLayer
from zope.publisher.interfaces.browser import IBrowserPage
from zope.traversing.interfaces import ITraversable
from z3c.form.interfaces import IFormLayer


class IFlourishLayer(IFormLayer):
    """SchoolTool flourish skin."""


class IFlourishBrowserLayer(IFlourishLayer, IDefaultBrowserLayer):
    """SchoolTool flourish skin."""


class IContentProviders(ITraversable):
    """Traversable collection of content providers."""


class IContentProvider(
    IBrowserPage, zope.contentprovider.interfaces.IContentProvider):

    view = zope.interface.Attribute(
        """The view the provider appears in.

        The view is the third discriminator of the content provider. It allows
        that the content can be controlled for different views.

        Having it stored as the parent is also very important for the security
        context to be kept.
        """)

    def __call__(*args, **kw):
        """Compute the response body."""


class IViewletOrder(Interface):
    """Attributes to specify viewlet display order."""

    after = Attribute(
        u"Display this viewlet after the given list of viewlets.")

    before = Attribute(
        u"Display this viewlet before the given list of viewlets.")

    requires = Attribute(
        u"Display only if all specified viewlets are available.")


class IViewlet(IContentProvider,
               ILocation,
               IBrowserPage,
               IViewletOrder):
    """A viewlet."""

    manager = zope.interface.Attribute(
        """The Viewlet Manager

        The viewlet manager for which the viewlet is registered. The viewlet
        manager will contain any additional data that was provided by the
        view, for example the TAL namespace attributes.
        """)


class IViewletManager(zope.viewlet.interfaces.IViewletManager,
                      ILocation,
                      IContentProvider):
    """A viewlet manager."""

    def collect():
        """Collect the viewlets and deduce their order."""


class IManagerViewlet(IViewlet, IViewletManager):
    """A viewlet that is also a manager."""


class IActiveViewletName(Interface):
    """Interface for adapter that returns the active viewlet name."""


class IPageBase(IBrowserPage, ILocation):

    template = Attribute(
        """Main page template, often used to render the page""")

    def update(self):
        pass

    def render(*args, **kw):
        pass


class IPage(IPageBase):
    title = zope.schema.TextLine(
        title=u"Page title", required=False)

    subtitle = zope.schema.TextLine(
        title=u"Page subtitle", required=False)

    page_template = Attribute(
        u"Template that renders all contents between header and footer.")

    content_template = Attribute(
        u"Template that renders the main content.")


IPage.setTaggedValue('flourish.template_content_type', 'html')


class IRMLTemplated(Interface):

    template = Attribute("""The template, renders RML""")

    def render(*args, **kw):
        """Render the RML."""


IRMLTemplated.setTaggedValue('flourish.template_content_type', 'xml')


class IPDFPage(IRMLTemplated, IPageBase):

    title = zope.schema.TextLine(
        title=u"PDF title", required=False)

    author = zope.schema.TextLine(
        title=u"PDF author", required=False)

    filename = zope.schema.TextLine(
        title=u"PDF file name", required=False)

    inline = zope.schema.Bool(
        title=u"Render inline", required=False)

    page_size = zope.schema.Tuple(
        title=u"Page size",
        value_type = zope.schema.Float(title=u"Size in pt (1/72 inch)"),
        required=False
        )

    rotation = zope.schema.Float(
        title=u"Page rotation",
        required=False
        )

    render_invariant = zope.schema.Bool(
        title=u"Render invariant",
        description=u"Render without influence form environment, like current time.",
        required=False)

    render_debug = zope.schema.Bool(
        title=u"Render debug",
        description=u"Render with debug information.",
        required=False)

    inline = zope.schema.Bool(
        title=u"Render inline", required=False)

    content_template = Attribute(
        u"Template that renders the main content.")


IPDFPage.setTaggedValue('flourish.template_content_type', 'xml')


class IPDFPart(IRMLTemplated, IViewlet):
    """A viewlet for RML."""


IPDFPart.setTaggedValue('flourish.template_content_type', 'xml')


class ITemplateSlots(ILocation):

    context = Attribute("The context object the view renders")
    request = Attribute("The request object driving the view")
    view = Attribute("The view")
    template = Attribute("A viewlet that renders a PDF <template>")


class IPageTemplate(IViewlet):

    slots = zope.schema.Object(
        title=u"Slot data",
        schema=ITemplateSlots,
        required=True)

    slots_interface = zope.schema.Object(
        title=u"Slot inteface",
        schema=zope.interface.interfaces.ISpecification,
        required=False)


class IFromPublication(IPublishTraverse):

    fromPublication = zope.schema.Bool(
        title=u'Accessed from publication',
        description=u'A flag that specifies if this part was built '
                     'by publication.',
        default=False,
        required=False)


class IAJAXParts(IViewletManager, IFromPublication):
    """AJAX parts of the view."""


class IAJAXPart(IViewlet, IFromPublication):
    """An ajax part of a view."""

    ignoreRequest = zope.schema.Bool(
        title=u'Ignore request',
        description=u'A flag that specifies if the request should be '
                     'regarded as unreliable (for example, part '
                     'was not obtained via publication).',
        default=True,
        required=False)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Person implementation and support objects
"""

import hashlib
from persistent import Persistent

from zope.interface import implements, implementer
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.container import btree
from zope.container.contained import Contained
from zope.component import adapts, adapter

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.app.membership import URIMembership, URIMember, URIGroup
from schooltool.app.overlay import OverlaidCalendarsProperty
from schooltool.person import interfaces
from schooltool.relationship import RelationshipProperty
from schooltool.securitypolicy.crowds import Crowd, AdministratorsCrowd
from schooltool.securitypolicy.crowds import ClerksCrowd, ManagersCrowd
from schooltool.person.interfaces import IPerson
from schooltool.app.security import ICalendarParentCrowd
from schooltool.securitypolicy.crowds import ConfigurableCrowd
from schooltool.person.interfaces import IPersonPreferences
from schooltool.person.interfaces import IPasswordWriter
from schooltool.securitypolicy.crowds import OwnerCrowd, AggregateCrowd


class PersonContainer(btree.BTreeContainer):
    """Container of persons."""

    implements(interfaces.IPersonContainer, IAttributeAnnotatable)

    super_user = None

    def __setitem__(self, key, person):
        """See `IWriteContainer`

        Ignores `key` and uses `person.username` as the key.
        """
        key = person.username
        btree.BTreeContainer.__setitem__(self, key, person)


class Person(Persistent, Contained):
    """Person."""

    implements(interfaces.IPersonContained, IAttributeAnnotatable)

    photo = None
    username = None
    _hashed_password = None

    groups = RelationshipProperty(URIMembership, URIMember, URIGroup)
    overlaid_calendars = OverlaidCalendarsProperty()

    def __init__(self, username=None, title=None):
        self.title = title
        self.username = username

    def setPassword(self, password):
        self._hashed_password = hash_password(password)

    def checkPassword(self, password):
        return (self._hashed_password is not None
                and hash_password(password) == self._hashed_password)

    def hasPassword(self):
        return self._hashed_password is not None

    def __eq__(self, other):
        if not IPerson.providedBy(other):
            return False
        return self.username == other.username


def hash_password(password):
    r"""Compute a SHA-1 hash of a given password.

        >>> hash_password('secret')
        '\xe5\xe9\xfa\x1b\xa3\x1e\xcd\x1a\xe8Ou\xca\xaaGO:f?\x05\xf4'

    Passwords should be ASCII or Unicode strings.

        >>> hash_password('\u263B')
        '\xe4\x13\xef\x8dv3\xba*P\xbb1\xa2k\x9c|,n\xe3mL'

    To avoid problems with a multitude of 8-bit encodings, they are forbidden

        >>> hash_password('\xFF') # doctest: +ELLIPSIS
        Traceback (most recent call last):
          ...
        UnicodeDecodeError: 'ascii' codec can't decode byte 0xff in ...

    None means "no password set, account is locked":

        >>> hash_password(None) is None
        True

    Security note: passwords are not salted, so it is possible to detect
    users that have the same password.
    """
    if password is None:
        return None
    return hashlib.sha1(password.encode('UTF-8')).digest()


def personAppCalendarOverlaySubscriber(person, event):
    """Add application calendar to overlays of all new persons.
    """
    app = ISchoolToolApplication(None, None)
    if app is None:
        # If we get this we are probably in the initial new-site setup
        # or creating a new manager during startup.  This should be
        # safe to ignore since it will happen very infrequently
        # (perhaps only once) and the manager can easily add the site
        # calendar to his/her overlay in the overlay selection view.
        return
    person.overlaid_calendars.add(ISchoolToolCalendar(app))


from schooltool.app.app import InitBase
class PersonInit(InitBase):

    def __call__(self):
        self.app['persons'] = PersonContainer()


class PublicCalendarCrowd(Crowd):
    def contains(self, principal):
        cal_public = IPersonPreferences(self.context).cal_public
        return cal_public


class PersonCalendarCrowd(AggregateCrowd):
    adapts(IPerson)
    implements(ICalendarParentCrowd)

    def crowdFactories(self):
        return [PublicCalendarCrowd, ManagersCrowd, ClerksCrowd,
                AdministratorsCrowd, OwnerCrowd]


def getCalendarOwner(calendar):
    return IPerson(calendar.__parent__, None)


@adapter(IPerson)
@implementer(interfaces.IPersonContainer)
def getContainerOfPerson(person):
    return person.__parent__


class PersonPasswordWriter(object):
    """Adapter of person to IPasswordWriter."""
    adapts(IPerson)
    implements(IPasswordWriter)

    def __init__(self, person):
        self.person = person

    def setPassword(self, password):
        """See IPasswordWriter."""
        self.person.setPassword(password)


class PersonListViewersCrowd(ConfigurableCrowd):
    """The crowd of people who can view the person list."""

    setting_key = 'everyone_can_view_person_list'


class PasswordWriterCrowd(ConfigurableCrowd):

    setting_key = 'persons_can_change_their_passwords'

    def contains(self, principal):
        from schooltool.app.browser import same # XXX
        app = ISchoolToolApplication(None)
        super_user = app['persons'].super_user
        if self.context.person is super_user:
            person = IPerson(principal, None)
            return same(person, super_user)
        if (ManagersCrowd(self.context.person).contains(principal) or
            ClerksCrowd(self.context.person).contains(principal)):
            return True
        return (ConfigurableCrowd.contains(self, principal) and
                OwnerCrowd(self.context.person).contains(principal))



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.app.
"""
import unittest
import doctest
from pprint import pprint

from zope.interface import implements
from zope.interface.verify import verifyObject
from zope.app.testing import setup
from zope.component import getMultiAdapter, provideAdapter
from zope.component import provideHandler
from zope.publisher.browser import TestRequest
from zope.security.interfaces import ISecurityPolicy
from zope.security.checker import defineChecker, CheckerPublic, NamesChecker
from zope.security.management import setSecurityPolicy
from zope.security.management import newInteraction, endInteraction
from zope.security.management import restoreInteraction

from schooltool.skin.flourish import interfaces
from schooltool.skin.flourish.viewlet import Viewlet, ViewletManager
from schooltool.skin.flourish.viewlet import ViewletProxy
from schooltool.skin.flourish.viewlet import ManagerViewlet


class TestViewlet(Viewlet):
    status = 'A fresh'

    def update(self):
        print 'Updating', self
        self.status = 'An updated'
        Viewlet.update(self)

    def render(self, *args, **kw):
        passed = (
            [repr(a) for a in args] +
            ['%s=%r' % a for a in sorted(kw.items())])
        return '%s ViewletTest. Called render(%s)' % (
            self.status, ', '.join(passed))

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.__name__)


def viewletClass(permission=CheckerPublic, **classdict):
    cls = type('TestViewlet', (TestViewlet, ), classdict)
    defineChecker(cls, NamesChecker(list(interfaces.IViewlet), permission))
    return cls


class TestManager(ViewletManager):

    def render(self, *args, **kw):
        rendered = ['TestManager.render:']
        rendered += [str(v.render(*args, **kw))
                     for v in self.viewlets]
        return '\n'.join(rendered)

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.__name__)


class TestManagerViewlet(ManagerViewlet):
    status = 'A fresh'

    def update(self):
        print 'Updating', self
        self.status = 'An updated'
        ManagerViewlet.update(self)

    def render(self, *args, **kw):
        passed = (
            [repr(a) for a in args] +
            ['%s=%r' % a for a in sorted(kw.items())])
        call_str = '%s %s.render(%s)' % (
            self.status, self.__class__.__name__,
            ', '.join(passed))
        rendered = [call_str]
        rendered += [str(v.render(*args, **kw))
                     for v in self.viewlets]
        return '\n'.join(rendered)

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.__name__)


def provideViewlet(factory, manager, name):
    provideAdapter(
        factory,
        (manager.context.__class__, manager.request.__class__,
         manager.view.__class__, manager.__class__),
        interfaces.IViewlet,
        name)


def doctest_Viewlet():
    """Tests for Viewlet.

        >>> viewlet = Viewlet('context', 'request', 'view', 'manager')

        >>> verifyObject(interfaces.IViewlet, viewlet)
        True

        >>> viewlet.context, viewlet.request, viewlet.view, viewlet.manager
        ('context', 'request', 'view', 'manager')

        >>> viewlet.__parent__
        'manager'

        >>> print viewlet.__name__
        None

        >>> viewlet.before, viewlet.after, viewlet.requires
        ((), (), ())

        >>> print viewlet.update()
        None

        >>> print viewlet.render()
        Traceback (most recent call last):
        ...
        NotImplementedError: `render` method must be implemented by subclass.


    """


def doctest_Viewlet_call():
    """Tests for Viewlet.__call__

        >>> viewlet = TestViewlet(None, 'request', None, None)

        >>> print viewlet.render()
        A fresh ViewletTest. Called render()

        >>> result =  viewlet('arg', option='something')
        Updating <TestViewlet None>

        >>> print result
        An updated ViewletTest. Called render('arg', option='something')

    Viewlets are not updated twice.

        >>> print viewlet('arg', option='something')
        An updated ViewletTest. Called render('arg', option='something')

    """


def doctest_ViewletProxy():
    """Tests for ViewletProxy.

        >>> print ViewletProxy.__doc__
        A viewlet proxy that turns a zope viewlet into flourish viewlet.

    Say, we have an old zope viewlet.

        >>> from zope.contentprovider.interfaces import IBeforeUpdateEvent
        >>> from zope.viewlet.viewlet import SimpleViewletClass

        >>> def beforeUpdate(e):
        ...     print 'About to update', e.object
        >>> provideHandler(beforeUpdate, [IBeforeUpdateEvent])

        >>> class ViewletBase(object):
        ...     def update(self):
        ...         print 'Updating.'

        >>> ZopeViewlet = SimpleViewletClass('simple_viewlet.pt',
        ...                                  bases=(ViewletBase,))

        >>> zope_viewlet = ZopeViewlet(
        ...     'context', TestRequest(), 'view', 'manager')

    It does not implement flourish viewlet interface.

        >>> verifyObject(interfaces.IViewlet, zope_viewlet)
        Traceback (most recent call last):
        ...
        DoesNotImplement: An object does not implement interface
        <InterfaceClass schooltool.skin.flourish.interfaces.IViewlet>

        >>> print interfaces.IViewlet(zope_viewlet, None)
        None

    But if we provide ViewletProxy adapter, we can adapt the old viewlet:

        >>> provideAdapter(ViewletProxy)

        >>> proxied = interfaces.IViewlet(zope_viewlet, None)

        >>> verifyObject(interfaces.IViewlet, proxied)
        True

    Proxied viewlet behaves as any other flourish viewlet.

        >>> proxied.context, proxied.request, proxied.view, proxied.manager
        ('context', <...TestRequest ...>, 'view', 'manager')

        >>> proxied.__parent__
        'manager'

        >>> print proxied.__name__
        <BLANKLINE>

        >>> proxied.before, proxied.after, proxied.requires
        ((), (), ())

     Let's render it.

        >>> result = proxied()
        About to update <...SimpleViewletClass from simple_viewlet.pt ...>
        Updating.

     Note that template of the old zope's viewlet knows nothing about
     methods and attributes of our proxy.  Hence the difference between
     __parent__s.

        >>> proxied.__parent__
        'manager'

        >>> print result
        <div>
          class: SimpleViewletClass from simple_viewlet.pt
          __parent__: view
          Class dict:
          ('_parent', 'view')
          ('context', 'context')
          ('manager', 'manager')
          ('request', <zope.publisher.browser.TestRequest instance
                       URL=http://127.0.0.1>)
        </div>

    Proxy has slots for attributes missing in Zope's viewlet.  We can
    change them if we like.

        >>> ViewletProxy.__slots__
        ('before', 'after', 'requires', '_updated')

        >>> print proxied._updated
        True

        >>> proxied.before = ('before',)
        >>> proxied.after = ('after',)
        >>> proxied.requires = ('requires',)

        >>> proxied.before, proxied.after, proxied.requires
        (('before',), ('after',), ('requires',))

        >>> proxied.view = 'new view'
        >>> proxied.manager = 'new manager'
        >>> proxied.context = 'new context'

        >>> proxied.context, proxied.request, proxied.view, proxied.manager
        ('new context', <...TestRequest ...>, 'new view', 'new manager')

        >>> proxied.__parent__
        'new manager'

    We can also add random new attributes as for normal class instances.

        >>> proxied.new_attribute = 'a new attribute'

    Viewlet does not get updated twice when rendering again:

        >>> result = proxied()

    Attributes from slots are not reflected in original viewlet's class dict.
    Otherwise, relevant changes made to proxied viewlet are transparent:

        >>> print result
        <div>
          class: SimpleViewletClass from simple_viewlet.pt
          __parent__: new view
          Class dict:
          ('_parent', 'new view')
          ('context', 'new context')
          ('manager', 'new manager')
          ('new_attribute', 'a new attribute')
          ('request', <zope.publisher.browser.TestRequest instance
                       URL=http://127.0.0.1>)
        </div>

    """


def doctest_ViewletManager():
    """Tests for ViewletManager.

        >>> context = 'context'
        >>> request = TestRequest()
        >>> view = 'view'

        >>> manager = ViewletManager(context, request, view)

        >>> verifyObject(interfaces.IViewletManager, manager)
        True

    Let's provide some viewlets for the manager.

        >>> provideViewlet(viewletClass(), manager, 'v1')
        >>> provideViewlet(viewletClass(), manager, 'v2')

        >>> getMultiAdapter(
        ...     (manager.context, manager.request,
        ...      manager.view, manager),
        ...     interfaces.IViewlet, 'v1')
        <TestViewlet None>

    Note that name of adapter gets assigned to viewlet.__name__.

        >>> manager.update()
        Updating <TestViewlet u'v1'>
        Updating <TestViewlet u'v2'>

        >>> pprint(manager.cache)
        {u'v1': <TestViewlet u'v1'>,
         u'v2': <TestViewlet u'v2'>}

        >>> manager.order
        [u'v1', u'v2']

        >>> manager.viewlets
        [<TestViewlet u'v1'>, <TestViewlet u'v2'>]

        >>> 'v1' in manager
        True

        >>> manager.get('v1')
        <TestViewlet u'v1'>

        >>> manager.get('v3', 'nosuchthing')
        'nosuchthing'

        >>> manager['v2']
        <TestViewlet u'v2'>

        >>> manager['v3']
        Traceback (most recent call last):
        ...
        KeyError: 'v3'

    """


def doctest_ViewletManager_call():
    """Tests for ViewletManager.

        >>> from zope.contentprovider.interfaces import IBeforeUpdateEvent

        >>> context = 'context'
        >>> request = TestRequest()
        >>> view = 'view'

        >>> manager = TestManager(context, request, view)

    Let's provide some viewlets for the manager.

        >>> provideViewlet(viewletClass(), manager, 'v1')
        >>> provideViewlet(viewletClass(), manager, 'v2')

        >>> def beforeUpdate(e):
        ...     print 'About to update', e.object
        >>> provideHandler(beforeUpdate, [IBeforeUpdateEvent])

    And now, call the manager.

        >>> result = manager('foo', bar='bar')
        About to update <TestManager None>
        About to update <TestViewlet u'v1'>
        Updating <TestViewlet u'v1'>
        About to update <TestViewlet u'v2'>
        Updating <TestViewlet u'v2'>

        >>> print result
        TestManager.render:
        An updated ViewletTest. Called render('foo', bar='bar')
        An updated ViewletTest. Called render('foo', bar='bar')

    """


def doctest_ViewletManager_collect():
    """Tests for ViewletManager.collect.

        >>> context = 'context'
        >>> request = TestRequest()
        >>> view = 'view'

        >>> manager = TestManager(context, request, view)

        >>> provideViewlet(
        ...     viewletClass(before=('v3', )),
        ...     manager, 'v1')

        >>> provideViewlet(viewletClass(), manager, 'v2')

        >>> provideViewlet(
        ...     viewletClass(after=('v4', )),
        ...     manager, 'v3')

        >>> provideViewlet(viewletClass(), manager, 'v4')

        >>> print manager.cache
        None

        >>> print manager.order
        None

    Collect the viewlets.

        >>> manager.collect()

        >>> pprint(manager.cache)
        {u'v1': <TestViewlet u'v1'>,
         u'v2': <TestViewlet u'v2'>,
         u'v3': <TestViewlet u'v3'>,
         u'v4': <TestViewlet u'v4'>}

        >>> manager.order
        [u'v2', u'v1', u'v4', u'v3']

    """


def doctest_ViewletManager_filter():
    """Tests for ViewletManager.filter

        >>> context = 'context'
        >>> request = TestRequest()
        >>> view = 'view'

        >>> class TestManager(ViewletManager):
        ...     def filterViewlets(self, viewlets):
        ...         print 'Filter viewlets:'
        ...         print sorted(viewlets)
        ...         return ViewletManager.filterViewlets(self, viewlets)

        >>> manager = TestManager(context, request, view)

        >>> provideViewlet(viewletClass(), manager, 'v1')

        >>> provideViewlet(
        ...     viewletClass(permission='deny'),
        ...     manager, 'v2')

        >>> provideViewlet(
        ...     viewletClass(requires=('v1',)),
        ...     manager, 'v3')

        >>> provideViewlet(
        ...     viewletClass(requires=('v2',)),
        ...     manager, 'v4')

        >>> provideViewlet(
        ...     viewletClass(requires=('v1', 'v3')),
        ...     manager, 'v5')

        >>> provideViewlet(
        ...     viewletClass(requires=('v1', 'v2')),
        ...     manager, 'v6')

    Let's create the manager.

        >>> manager = TestManager(context, request, view)

    Filters are applied when collecting viewlets.

        >>> viewlets = list(manager.viewlets)
        Filter viewlets:
        [(u'v1', <TestViewlet u'v1'>),
         (u'v2', <TestViewlet u'v2'>),
         (u'v3', <TestViewlet u'v3'>),
         (u'v4', <TestViewlet u'v4'>),
         (u'v5', <TestViewlet u'v5'>),
         (u'v6', <TestViewlet u'v6'>)]

    Since viewlets are cached, they will not be filtered again on update:

        >>> manager.update()
        Updating <TestViewlet u'v1'>
        Updating <TestViewlet u'v3'>
        Updating <TestViewlet u'v5'>

        >>> print manager.viewlets
        [<TestViewlet u'v1'>, <TestViewlet u'v3'>, <TestViewlet u'v5'>]

    Lets try the default process and go straight to update.

        >>> manager = TestManager(context, request, view)
        >>> manager.update()
        Filter viewlets:
        [(u'v1', <TestViewlet u'v1'>),
         (u'v2', <TestViewlet u'v2'>),
         (u'v3', <TestViewlet u'v3'>),
         (u'v4', <TestViewlet u'v4'>),
         (u'v5', <TestViewlet u'v5'>),
         (u'v6', <TestViewlet u'v6'>)]
        Updating <TestViewlet u'v1'>
        Updating <TestViewlet u'v3'>
        Updating <TestViewlet u'v5'>

    """


def doctest_ViewletManager_presort():
    """Tests for ViewletManager.presort

        >>> context = 'context'
        >>> request = TestRequest()
        >>> view = 'view'

        >>> class StubViewlet(object):
        ...     def __init__(self, name, before=(), after=()):
        ...         self.__name__ = name
        ...         self.before = before
        ...         self.after = after

        >>> manager = ViewletManager(context, request, view)

        >>> viewlets = {
        ...     'a_last': StubViewlet('', after=('v_a', '*')),
        ...     'v_a': StubViewlet('v_a', after=('v_c',)),
        ...     'v_b': StubViewlet('v_b', before=('v_a',)),
        ...     'z_first': StubViewlet('z_first', before=('*',),
        ...                                       after=('two',)),
        ...     'v_c': StubViewlet('v_c', before=('*',), after=('*',)),
        ...     }

    Default presort returns viewlets sorted by name.

        >>> manager.presort(viewlets)
        ['a_last', 'v_a', 'v_b', 'v_c', 'z_first']

    """


def doctest_ManagerViewlet():
    """Tests for ManagerViewlet.

        >>> names = sorted(interfaces.IManagerViewlet)

        >>> context = 'context'
        >>> request = TestRequest()
        >>> view = 'view'
        >>> manager = 'manager'

        >>> viewlet = ManagerViewlet(context, request, view, manager)
        >>> verifyObject(interfaces.IManagerViewlet, viewlet)
        True

        >>> viewlet.context, viewlet.request, viewlet.view, viewlet.manager
        ('context', <...TestRequest ...>, 'view', 'manager')

        >>> viewlet.__parent__
        'manager'

        >>> print viewlet.update()
        None

        >>> print viewlet.render()
        Traceback (most recent call last):
        ...
        NotImplementedError: `render` method must be implemented by subclass.

    The render method is inherited from the Viewlet.

        >>> viewlet.__class__.render == Viewlet.render
        True

    """


def doctest_ManagerViewlet_call():
    """Tests for ViewletManager.__call__

        >>> v = TestManagerViewlet(None, 'request', None, None)

        >>> print v.render()
        A fresh TestManagerViewlet.render()

        >>> result =  v('arg', option='something')
        Updating <TestManagerViewlet None>

        >>> print result
        An updated TestManagerViewlet.render('arg', option='something')

    """


def doctest_ManagerViewlet_viewlets():
    """Tests for ViewletManager.

        >>> from zope.contentprovider.interfaces import IBeforeUpdateEvent

        >>> context = 'context'
        >>> request = TestRequest()
        >>> view = 'view'
        >>> manager = 'manager'

        >>> viewlet = TestManagerViewlet(context, request, view, manager)

    Let's provide some viewlets for the manager.

        >>> provideViewlet(viewletClass(), viewlet, 'v1')
        >>> provideViewlet(viewletClass(), viewlet, 'v2')

        >>> def beforeUpdate(e):
        ...     print 'About to update', e.object
        >>> provideHandler(beforeUpdate, [IBeforeUpdateEvent])

        >>> print viewlet.cache
        None

        >>> print viewlet.viewlets
        [<TestViewlet u'v1'>, <TestViewlet u'v2'>]

        >>> pprint(viewlet.cache)
        {u'v1': <TestViewlet u'v1'>, u'v2': <TestViewlet u'v2'>}

        >>> print viewlet.viewlets
        [<TestViewlet u'v1'>, <TestViewlet u'v2'>]

    Let's render:

        >>> result = viewlet('foo', bar='bar')
        About to update <TestManagerViewlet None>
        Updating <TestManagerViewlet None>
        About to update <TestViewlet u'v1'>
        Updating <TestViewlet u'v1'>
        About to update <TestViewlet u'v2'>
        Updating <TestViewlet u'v2'>

        >>> print result
        An updated TestManagerViewlet.render('foo', bar='bar')
        An updated ViewletTest. Called render('foo', bar='bar')
        An updated ViewletTest. Called render('foo', bar='bar')

    """


class SecurityPolicy(object):
    implements(ISecurityPolicy)

    def checkPermission(self, permission, object):
        return permission == 'allow'


def setUp(test=None):
    setup.placelessSetUp()
    test.globs['__policy'] = setSecurityPolicy(SecurityPolicy)
    endInteraction()
    newInteraction()


def tearDown(test=None):
    setSecurityPolicy(test.globs['__policy'])
    restoreInteraction()
    setup.placelessTearDown()


def test_suite():
    optionflags = (doctest.ELLIPSIS |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_NDIFF)

    return unittest.TestSuite([
                doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                     optionflags=optionflags),
           ])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')
